
# Perl file to be required (general goodies)

###############################################
# Return true if a core exists (and remove it)

sub core_exist {

    if (-f 'core'){
        unlink('core');
        return 1;
    }
    return 0;
}

############################
#
# Return a name for lame signal numbers
#

@sig_name = (
    'NONE',
    'SIGHUP  - hangup',
    'SIGINT  - interrupt',
    'SIGQUIT - quit',
    'SIGILL  - illegal instruction',
    'SIGTRAP - trace trap',
    'SIGIOT  - IOT (aka abort)',
    'SIGEMT  - EMT instruction',
    'SIGFPE  - floating point exception',
    'SIGKILL - kill',
    'SIGBUS  - bus error',
    'SIGSEGV - segmentation violation',
    'SIGSYS  - bad argument to system call',
    'SIGPIPE - broken pipe',
    'SIGALRM - alarm',
    'SIGTERM - termination'
);

sub name_sig {

    local($sig) = @_;

    if ( $sig <= 15 ){
        return $sig_name[$sig];
    } else {
        return "machine specific signal";
    }
}

############################
#
# Do a make of these targets in the pgms directory
#

sub make {

    local(@targets) = @_;
    local($target,$ret,$err);

    $ret = 1;

    chdir 'pgms' || return 0;

    foreach $target (@targets){

        $err = system("make $target > make.err 2>&1");

        if (open(MKE_ERR,'make.err')){
            print <MKE_ERR>;
            close MKE_ERR;
            unlink 'make.err';
        }

        $err /= 256;
        if ($err){
            print "Make failed for target $target.\n";
            $ret = 0;
        }
    }

    chdir '..';
    return $ret;
}
############################

sub get_dir {

   local($dir,$pattern) = @_;
   local(@flist);

   @flist = ();

   # If no pattern supplied, return everything except '.' and '..'
   $pattern = '!/^\.\.?$/' if (! $pattern);

   opendir(DIR, "$dir");

   eval "\@flist = grep($pattern, sort(readdir(DIR)))";
   closedir( DIR );

   return @flist;
}

############################
#
#

sub set_max_time {

    local($secs) = @_;

    $SIG{'ALRM'} = 'alrm_handler';

    alarm $secs;
}

############################

sub alrm_handler {

    print "\n\nCoral TestSuite $suite has run out of time (> $max_time) - quitting.\n\n";

    # Try to clean up a little
    
    &core_exist;

    if ((-d 'pgms') && (-f 'pgms/Makefile')){
        chdir('pgms');
        &make('clean');
    }

    exit 1;
}

############################

sub allow_quit {

     # Watch for ^C and other requests to quit.
    $SIG{'HUP'}  = 'quit_handler';
    $SIG{'INT'}  = 'quit_handler';
    $SIG{'TERM'} = 'quit_handler';
}

############################

sub quit_handler {

    print "\n\nSignal requesting quit recieved. - quitting.\n\n";

    # Try to clean up a little
    
    &core_exist;

    if ((-d 'pgms') && (-f 'pgms/Makefile')){
        chdir('pgms');
        &make('clean');
    }

    exit 0;
}

############################

# So that require succeed.
1;
