/*****************************************************************************/
//	Arg.h
/*****************************************************************************/

#ifndef _Arg_H_
#define _Arg_H_

#include <stdio.h>
#include <stdlib.h>

#ifdef NULL
#  undef NULL
#  define NULL (void *) 0
#endif

/*****************************************************************************/

struct corVarLink;
struct corTermLink;
class corTuple;
struct corVarArg;
typedef int Variable;		   // This is distinct from VarArg !!
typedef int VarFunc(corVarArg* var, void* data);

class corTerm;
class corBindEnv;
class corRelation;
struct corBitVector;

/*****************************************************************************/

typedef char arg_kind;
#define COR_NUM_CONST	 0
#define COR_VARIABLE	  1
#define COR_DONTCARE	  2
#define COR_FUNCTOR	   3
#define COR_CONST_ARG	 4
#define COR_SYMBOL_CONST  5
#define COR_RELATION		  6
#define COR_GROUPING	  7
#define COR_TUPLE_ARG	 8
#define COR_TUPLE_VAR	 9

/*****************************************************************************/
// Hashing and HashConsing constants:
// 		Any hash value other than VarHashValue and UnknownHashValue 
//				is a hash value of a ground term
//		A term whose hash value is VarHashValue is non-ground.
//		A term whose hash value is UnknownHashValue may be ground or non-ground
//				-- its hash value has not been computed.
// 		When hashconsing is used (USE_HASHCONSING is set), FuncArg.hash()
//				is either VarHashValue, or is a pointer to a
//				constant term that is equal to (FuncArg, bindenv) 
//				(the pointer is cast to HashVal). 
//				CAVEAT:  We assume that the pointer cannot be one of 
//						VarHashValue or UnknownHashValue.
//

#define VarHashValue ( (HashVal) -1)

// This value is not generally supported!
#define UnknownHashValue ( (HashVal) -2)

typedef unsigned int HashVal;

inline int corisGroundHashValue( HashVal i)
{
	return (i != VarHashValue && i != UnknownHashValue);
}

inline int corIntToHash(int i)
{
	return ((HashVal)(-1) >> 1) & i;
}

//oArg: old Arg

class corCArg;

typedef class corArg {
public:
	arg_kind  _kindof;
	virtual corVarLink *var_list(corVarLink* rest);
	virtual Variable max_vars();

	virtual void print(corBindEnv *context, FILE *file) const;
	virtual void print(corBindEnv *context, FILE *file, char *fmt_str) const;
	virtual void print_dump(corBindEnv *context, FILE *file) const
	  {print(context, file);};
 
	// sprint is used for printing on a string. The second
	// argument is advanced to the end of the written portion,
	// so that the next sprint can be called.
	virtual void sprint(char *str, int *pos, corBindEnv *context = NULL) const =0;
	void printon(FILE *file) const { print((corBindEnv*)0, file); }
	virtual void printon(FILE *file, char *fmt_str) const;
	virtual void output() const { print((corBindEnv*)0, stderr); }

	inline arg_kind kindof() const { return _kindof;}
	virtual int subkind() { return 0;}

	// Returns true if it is a NumArg or Symbol
	virtual int isConstant() = 0;
	virtual int equals(corArg *arg2);

	virtual HashVal hash(corBindEnv *context = 0);

	virtual corCArg simplify(corBindEnv *context, corTermLink *&renamed_vars,
						corBindEnv *const_env, corBindEnv *rename_only_env);

	// For each VarArg* 'v' in 'this', call 'func(v, data)'.
	virtual int scan_vars(void *data, VarFunc *func) ;
	corArg() { _kindof = COR_CONST_ARG;}
	virtual ~corArg() {};
} *ArgPtr;

/*------------------------------------------------------------------
 *::sprint(char *string, int* pos, BindEnv *env)
		print the argument on string, starting at *pos.  Update *pos
		to be the position of the end of the printed argument.
 *::isConstant(): 
		Returns true if (this) has no variables in it, and returns 
		false otherwise.
 *::simplify(BindEnv *env, TermLink *& renamed_vars, BindEnv* dont_rename_env,
				BindEnv *rename_only_env)
		Renames the variables in (this) interpreted in env.  
		- renamed_vars is a list of variable names along with their new names
		(their position in the list + dont_rename_env_max_var gives the new
		name).
		- dont_rename_env is a bindenv, the variables in which need not be 
		renamed.  Any structure that is interpreted in dont_rename_env is left
		unchanged by simplify.  
		- rename_only_env -- if this is non-null, and an argument is found
				to be from a bindenv other than dont_rename_env and
				rename_only_env, simplify returns failure.
				If this is null, all bindenvs other than dont_rename_env are
				renamed.

		 Oddities/Quirks :: 

				WARNING:
				-  This function assumes that if rename_only_var is non null, 
						the function has been called recursively from backpatch_trail.
						In this case, there is no need to follow terms that are
						interpreted in const_env.

				- Variables of the form _X1, _X2, ...  should not be generated 
				elsewhere in the system.  They are reserved for this function to 
				generate.
 *::var_list(VarLink *rest)
		Return (reversed) list of all variables in 'this'. Append 'rest' 
		to the end.  Default: there are no variables in 'this'.
  *::max_vars()
		Returns 1 + the number of the highest numbered variable in (*this).
 *::equals(Arg *arg)
		Returns true if (this) is equal to arg, false otherwise.
		Note that no bindenv is provided, so this really works only
		for constant structures.
 *::hash(BindEnv *env)
		Compute the hash/hashcons value of (this) interpreted in env.
 *::scan_vars(void *data, VarFunc *func)
		Executes the specified function for each variable in (this).
		- this is the first argument of the function call.
		- data is the second argument of the function call.
 *::print(BindEnv *env, FILE *file)
		Interpret (*this) in env, and print it out on to the file.

 -------------------------------------------------------------------*/
/*****************************************************************************/

// NOTE: Do not edit this without also editing builtin_ops in Gram.y!
enum AggregateKind {
	SumAggregation,		// sum(<X>)
	ProductAggregation,	// prod(<X>)
	CountAggregation,	  // count(<X>)
	MinAggregation,		// min(<X>)
	MaxAggregation,		// max(<X>)
	AvgAggregation,		// avg(<X>)
	SumDAggregation,	   // sum_distinct(<X>)
	ProdDAggregation,	  // prod_distinct(<X>)
	CountDAggregation,	 // count_distinct(<X>)
	AvgDAggregation,	   // avg_distinct(<X>)
	SetAggregation,		// distinct(<X>)
	AnyAggregation,		// any(<X>)
	NoAggregation, // create_functor in Gram.y assumes this is last.
};

/*****************************************************************************/

/* Copy and simplify from old_args to new_args. */
// WARNING:  may need to change NULL below to UnusedEnv

class corArgList;
class corCArg;

extern int corCopyArgs(corArgList * new_args, corArgList *old_args, corBindEnv *env,
	int arity, corBindEnv *dont_rename_env, corBindEnv *rename_only_env = NULL);
extern int corCopyArg(corCArg& new_arg,  corCArg& old_arg, corBindEnv *env);

/*****************************************************************************/

#endif _Arg_H_



















