#ifndef _CARG_H
#define _CARG_H
//CArg.h
//the arg that is more efficient

#include "Arg.h"
#include "NumArg.h"
#include "SymbolArg.h"
#include "CoralError.h"

enum CArgKind {EMPTY_ARG, INT_ARG, DOUBLE_ARG, SYMBOL_ARG, OTHER_ARG, NOT_A_NUMBER};
#undef FALSE
#undef TRUE
enum {FALSE, TRUE};

// EMPTY_ARG: right after initialization, no real value
// INT_ARG, DOUBLE_ARG, SYMBOL_ARG: a number or a string, directly stored
//        in the union
// OTHER_ARG:pointer to other arg type

class corNumArg;
class corGrouping;
class corStorageRelation;
class corHashSimpleRelation;
class corAvgArg;
class corDistArg;
class corADistArg;
class corArrayArg;
class corRelation;
class corFuncArg;
class corSymbol;
class corVarArg;
class corExprArg;

class corCArg {
  CArgKind my_arg_kind;
  union {
    double double_arg;
    int int_arg;
    char *symbol;
    corArg *other; 
   } ;
 public:
  corCArg(void) ;
  corCArg(corArg* arg) ;
  corCArg(const corCArg & a);
  corCArg(int n);
  corCArg(double d);
  void init();
  ~corCArg() {};
  corCArg& operator= (const corCArg& a);
  corCArg& operator= (int i);
  corCArg& operator= ( double d);
  corCArg& operator= (corArg * p) ;
  int operator ==(const corCArg& c) const;
  int operator ==(corArg* a) const;
  int operator !=(corArg* a) const;

  int IsEmpty() const;		// empty arg or Other_arg (0)

  int IsPtrEqual(const corCArg& c) const; // Number: equals, Other (ptr), 

  //test if two CArgs are equal,
  int equals(const corCArg& arg2) const;

  CArgKind Kind() const { return my_arg_kind ;};
  arg_kind kindof() const { 
    if ((my_arg_kind == OTHER_ARG) && (other))
      return other->kindof();
    else if( i_am_num() )
      return COR_NUM_CONST;
    else
      return COR_CONST_ARG;	// Arg() gives this value
   };

  int i_am_num() const{
    return(my_arg_kind == INT_ARG) || (my_arg_kind == DOUBLE_ARG);
   };

  //reports the number kind of a CArg, if it is a number
  numarg_kind cNum_kindof() const {
    if (my_arg_kind == INT_ARG) return COR_INTEGER;
    if (my_arg_kind == DOUBLE_ARG) return COR_DOUBLE;
    if ((my_arg_kind == OTHER_ARG) && (other->_kindof == COR_NUM_CONST))
      return ((corNumArg*) other)->num_kindof();
    CORAL_error(COR_FATAL_ERROR, "","cNum_kindof");
   };

  double NumDval()const {
    if (my_arg_kind == DOUBLE_ARG) return double_arg;
    if (my_arg_kind == INT_ARG) return (double)int_arg;
    CORAL_error(COR_FATAL_ERROR, "","NumDval");
   };

  int NumIval()const {
    if (my_arg_kind == INT_ARG) return int_arg;
    CORAL_error(COR_FATAL_ERROR, "","NumIval");
   };
 
  int isConstant() const{ 
    if (my_arg_kind == OTHER_ARG)
      return other->isConstant();
    else if (my_arg_kind !=EMPTY_ARG)
      return 1;
    else
      CORAL_error(COR_FATAL_ERROR, "","isConstant");
   };
 
  int is_int() const {
    return my_arg_kind == INT_ARG;
   };
  int is_short() const { return is_int();};
  int is_long() const { return is_int();};
  int is_float() const {
    return my_arg_kind == DOUBLE_ARG ;
   };
  int is_double() const {
    return my_arg_kind == DOUBLE_ARG;
   };
  
  // compares two numbers. assume they both are.
  int compare_num(const corCArg& arg2) const;
  
  HashVal hash(corBindEnv *context = 0) const;

  corCArg simplify(corBindEnv *context, corTermLink *&renamed_vars,
			corBindEnv *const_env, corBindEnv *rename_only_env) const;

  //returns the length of the string, if symbol arg. no err check
  int GetSymbolLen() const {
    return ((corSymbol*)other)->_length;
   }; 

  char* GetSymbolStr() const {
    return ((corSymbol*)other)->str;
   };

  corVarLink *var_list(corVarLink* rest);
  Variable max_vars() ;
  void print(corBindEnv *context, FILE *file) const;
  void print(corBindEnv *context, FILE *file, char *fmt_str) const ;
  void print_dump(corBindEnv *context, FILE *file) const ;
  void sprint(char *str, int *pos, corBindEnv *context = NULL) const;
  void printon(FILE *file) const;
  void printon(FILE *file, char *fmt_str) const;
  void output() const;
  int subkind() const;
  
  int scan_vars(void *data, VarFunc *func);

  corGrouping* GroupingP() const {
    return (corGrouping*) other;
   };

  corStorageRelation* StorageRelP() const {
    return (corStorageRelation*) other;
   };

  corHashSimpleRelation* HashSimpleRelP() const {
    return (corHashSimpleRelation*) other;
   };

  corAvgArg* AvgArgP() const {
    return (corAvgArg*) other;
   };

  corDistArg* DistArgP() const {
    return (corDistArg*) other;
   };

  corADistArg* ADistArgP() const {
    return (corADistArg*) other;
   };

  corArrayArg* ArrayArgP() const {
    return (corArrayArg*) other;
   };

  corRelation* RelationP() const {
    return (corRelation*) other;
   };

  corFuncArg* FuncArgP() const {
    return (corFuncArg*) other;
   };

  corSymbol* SymbolP() const {
    return (corSymbol*) other;
   };

  corVarArg* VarArgP() const {
    return (corVarArg*) other;
   };

  corExprArg* ExprArgP() const { return (corExprArg*) other; };
 };


inline char* corSymbolString(corCArg c) {
  return c.SymbolP()->str;
 };

inline char* corSymbolString(corCArg *c) {
  return c->GetSymbolStr();
 };


const corCArg corNullCArg;
const corCArg corZeroCArg(0);
const corCArg corMinusOneCArg(-1);
const corCArg corOneCArg(1);

#endif _CARG_H
