/*****************************************************************************/
//    GArrayBindEnv.C
/*****************************************************************************/

#include "GArrayBindEnv.h"
#include "Globals.h"

/****************************************************************************/
corGArrayBindEnv::corGArrayBindEnv(int size, int incr /* = 128 */)
{
    _limit = size;
    ASSERT((size >=0))
    bindarray = new corTerm_garray(size,incr); 
    _is_versioned = 0;
#    ifdef DEBUG
        corC_ConstructEnvCount++;
        corC_ConstructEnvSize += size;
#    endif
}

/****************************************************************************/
corGArrayBindEnv::corGArrayBindEnv()
{
    _limit = 0;
    bindarray = NULL;
    _is_versioned = 0;
#    ifdef DEBUG
        corC_ConstructEnvCount++;
#    endif
}

/****************************************************************************/
corGArrayBindEnv::~corGArrayBindEnv()
{
#    ifdef DEBUG
        corC_DestructEnvCount++;
        corC_DestructEnvSize += max_var_number();
#    endif
    delete bindarray;    // This delete can cause problems if
                // bindarrays are shared across bindenvs.
}

/****************************************************************************/
/*
    This function is quite efficient if the array has not been accessed
    at all.  Otherwise it may take as much time as the size of the bindenv.
    bindarray->get_next(i) scan fast through the array, when primed with 
    i=-1.  It skips large empty regions of the array very fast.
*/

int corGArrayBindEnv::all_free()
{
    int i;

    for(i=bindarray->get_next(-1); i < max_var_number() && i >= 0;
        i=bindarray->get_next(i))
    {
        if((*bindarray)[i].expr.IsEmpty() == 0)
            return(0);
    }
    return(1);
}

/****************************************************************************/
corBindEnv* corGArrayBindEnv::make_version()
{
    corGArrayBindEnv* copy = new corGArrayBindEnv();
    copy->_limit = _limit;
    copy->bindarray =  bindarray->make_copy();
#    ifdef DEBUG
        corC_ConstructEnvSize += _limit;
#    endif
    return(copy);
}


/****************************************************************************/
corBindEnv* corGArrayBindEnv::make_version_and_swap()
{
    // This is a hack --- see note on versioning-hack for more information
    fprintf(corexEnv->error_file,
      "System Error: GArrayBindEnv:make_version_and_swap - not yet implemented\n");
    return NULL;
}

/****************************************************************************/
void corGArrayBindEnv::restore(corBindEnv *)
{
    // This is a hack --- see note on versioning-hack for more information
    // env has to be an VersionedBindEnv
    fprintf(corexEnv->error_file,
        "System Error: GArrayBindEnv:restore - not yet implemented\n");
}

/****************************************************************************/
corBindEnv* corGArrayBindEnv::copy_shell()
{
    // This is a hack -- see note on versioning-hack for more information
    fprintf(corexEnv->error_file,
        "System Error: GArrayBindEnv:copy_shell - not yet implemented\n");
    return NULL;
}

/****************************************************************************/

