/*****************************************************************************/
//    BuiltinBinaryRelation.C
/*****************************************************************************/

#include "BuiltinBinaryRelation.h"

#include "InitBuiltins.h"
#include "Globals.h"
#include "TupleIterator.h"
#include "CoralError.h"

/*****************************************************************************/
corBuiltinBinaryRelation::corBuiltinBinaryRelation(int a, Name n, BinaryProc bin_op)
    : corBuiltinRelation(a, n, NULL)
{
    op = bin_op;
}

/*****************************************************************************/
corBindEnv * corBuiltinBinaryRelation::get_next(corTupleIterator& iterator)
{
    corBindEnv *env = iterator.bindenv;

    if (iterator.ipos > 0) {
        iterator.stack_mark.pop_to();
        iterator.set_no_match(); 
        return env;
    }
    iterator.ipos++;
    iterator.stack_mark.get_mark();

    // Term terms[3];
    for (int i = 0; i < 3; i++) {
        corstatic_terms[i].expr = iterator.arg_list[i],
            corstatic_terms[i].bindenv = env;
        FULL_DEREFERENCE_TERM(corstatic_terms[i]);
    }
    
#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file, "BuiltinRelation:"); 
            corstatic_terms[2].printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, " = %s (", corSymbolString(name));
            corstatic_terms[0].printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ",");
            corstatic_terms[1].printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, ")\n");
        }
#    endif

    //    Specialized for numeric binary operations
    //    WARNING: These operators work only one way currently. There should
    //    be different operators called depending on the binding pattern.

    if (corstatic_terms[0].kindof() != COR_NUM_CONST ||
        corstatic_terms[1].kindof() != COR_NUM_CONST)
    {
        CORAL_error(COR_NON_NUMERIC_ARGS, NULL, "get_next");
        fprintf(corexEnv->error_file,"(");
        corstatic_terms[0].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,",");
        corstatic_terms[1].printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,")\n");
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
        return env;
    }

    //Term result_term;
    corusable_terms[0].expr = (*op)(corstatic_terms[0].expr,
        corstatic_terms[1].expr);
    corusable_terms[0].bindenv = NULL;
    if (corusable_terms[0].expr.IsEmpty()) {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
        return env;
    }
    
    if (!corunify_args (corusable_terms[0], corstatic_terms[2])) {
        iterator.set_no_match();
        iterator.stack_mark.pop_to();
        return env;
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_get_next) {
            fprintf(corexEnv->trace_file,"Result = "); 
            corstatic_terms[2].printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "\n");
        }
#    endif

    return env;
}

/*****************************************************************************/

