//**************************************************************************
//    DBSolvers.C
//**************************************************************************

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <unistd.h>

#include "Globals.h"
#include "BuiltinRelation.h"
#include "BuiltinSymbols.h"
#include "CoralError.h"
#include "Util.h"
#include "Database.h"
#include "TupleIterator.h"
#include "ClientProtocol.h"
#include "CArg.h"

//**************************************************************************

        // For parsing from strings
#define BUFF_MAX 5000

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corCreateDBSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to ws_create needed", NULL);
        FAIL_RETURN
    }
    else {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        //FULL_DEREFERENCE_TERM (term1);
        cordeep_dereference_term (term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "CreateDBSolver");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            FAIL_RETURN
        }
        
        char *actual_string = corSymbolString(term1.expr);
        corcreateDB(corEnterSymbol(actual_string));
    }
    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corDestroyDBSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to ws_destroy needed", NULL);
        FAIL_RETURN
    }
    else {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        //FULL_DEREFERENCE_TERM (term1);
        cordeep_dereference_term (term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "DestroyDBSolver");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            FAIL_RETURN
        }

        char *actual_string = corSymbolString(term1.expr);
        cordestroyDB(corEnterSymbol(actual_string));
    }
    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corCurrentDBSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() == 0) {
        fprintf(corexEnv->error_file, "Current Workspace is: %s\n",
            corSymbolString(corCurDB->name));
        return(iterator.bindenv);
    }

    if (iterator.arg_list.count() != 1) {
        CORAL_error(COR_BAD_ARITY, "One argument to ws_current needed", NULL);
        FAIL_RETURN
    }
    else {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        cordeep_dereference_term (term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "CurrentDBSolver");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            FAIL_RETURN
        }

        char *actual_string = corSymbolString(term1.expr);
        corcurrentDB(corEnterSymbol(actual_string));
        if (!strcmp(corSymbolString(corCurDB->name), "default_ws")) 
            corparseEnv->prompt = coralConfig.prompt;
        else {
            strcpy(corparseEnv->prompt_name, corSymbolString(corCurDB->name));
            strcat(corparseEnv->prompt_name, coralConfig.prompt);
         }

        if (coralConfig.server_mode)
            fprintf(corexEnv->print_file,"%s%s\n",CP_PRMT,corparseEnv->prompt);

          // Update Current database for the current user.
        corcurUser->corCurDB = corCurDB;

    }
    return(iterator.bindenv);                
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv * corListDBSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 0) {
        CORAL_error(COR_BAD_ARITY, "no arguments to list_ws needed", NULL);
        FAIL_RETURN
    }

    corlistEveryDB();

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corSaveDBSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    char *actual_string1, *actual_string2;

    if ((iterator.arg_list.count() > 2) || (iterator.arg_list.count() < 1)) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to ws_save needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "SaveDBSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    if (iterator.arg_list.count() == 2) {
        corTerm term2 (iterator.arg_list[1], iterator.bindenv);
        //FULL_DEREFERENCE_TERM (term2);
        cordeep_dereference_term (term2);
        if (term2.expr.kindof() != COR_SYMBOL_CONST) {
            CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "SaveDBSolver");
            term2.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            FAIL_RETURN;
        }
        actual_string2 = corSymbolString(term2.expr);
    }
    else
        actual_string2 = NULL;

    actual_string1 = corSymbolString(term1.expr);

    FILE *fd;
    if (!(fd = fopen(actual_string1, "a"))) {
        CORAL_error(COR_CANNOT_OPEN_FILE, actual_string1, "SaveDBSolver");
        FAIL_RETURN        
    }

    corsaveDB(actual_string2 ? corEnterSymbol(actual_string1) : NULL, fd);
                
    fclose(fd);

    return(iterator.bindenv);                    
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corRenameDBSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    char *actual_string1, *actual_string2;

    if ((iterator.arg_list.count() < 1) || (iterator.arg_list.count() > 2)){
        CORAL_error(COR_BAD_ARITY, "2 arguments to ws_rename needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "RenameDBSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    if (iterator.arg_list.count() == 1) {
        actual_string1 = corSymbolString(term1.expr);
        correnameDB(corCurDB->name, corEnterSymbol(actual_string1));
        return(iterator.bindenv);        
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term2);
    cordeep_dereference_term (term2);
    if (term2.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "RenameDBSolver");
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }
    actual_string1 = corSymbolString(term1.expr);
    actual_string2 = corSymbolString(term2.expr);

    correnameDB(corEnterSymbol(actual_string1), corEnterSymbol(actual_string2));
    
    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corOpenRelSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if (iterator.arg_list.count() != 2) {
        CORAL_error(COR_BAD_ARITY, "2 arguments to rel_open needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a func-arg", "OpenRelSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term2);
    cordeep_dereference_term (term2);
    if (term2.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "OpenRelSolver");
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    char *actual_string2 = corSymbolString(term2.expr);

    coropenRelInCurDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2));
    
    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* corRelDisplay(corBuiltinRelation&, corTupleIterator& iterator)
{
    char name[100];
    char *tmp_file_name = cormk_tmp_file_name(name);
    FILE *tmp_file = coropen_file(tmp_file_name);
    if (!tmp_file) {
        FAIL_RETURN
    }

    if (iterator.arg_list.count() == 0) {
        corlistRelsInDB(corCurDB->name, tmp_file);
    }
    else if (iterator.arg_list.count() == 1) {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        //FULL_DEREFERENCE_TERM (term1);
        cordeep_dereference_term (term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST) {
            CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "RelDisplay");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            FAIL_RETURN
        }
        char *actual_string = corSymbolString(term1.expr);
        corlistRelsInDB(corEnterSymbol(actual_string), tmp_file);
    }
    else {
        fclose(tmp_file);
        CORAL_error(COR_BAD_ARITY, "0 or 1 argument to list_rels needed", NULL);
        FAIL_RETURN
    }

    fclose(tmp_file);

    char name1[100];
    char *tmp_file_name1 = cormk_tmp_file_name(name1);

    char buf[200];
    sprintf(buf, "sort +0 +0 %s > %s", tmp_file_name, tmp_file_name1);
    if (system(buf)) {
        corDisplayFile(tmp_file_name);
    }
    else {
        corDisplayFile(tmp_file_name1);
        unlink(tmp_file_name1);
    }

    unlink(tmp_file_name);
    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corCloseRelSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if ((iterator.arg_list.count() != 2) && (iterator.arg_list.count() != 1)) {
        CORAL_error(COR_BAD_ARITY, "1 or 2 arguments to rel_open needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a func-arg", "CloseRelSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    if (iterator.arg_list.count() == 1) {
        corremoveRelFromDB(term1.expr.FuncArgP(), corCurDB->name);
        return(iterator.bindenv);
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term2);
    cordeep_dereference_term (term2);
    if (term2.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "CloseRelSolver");
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }
    char *actual_string2 = corSymbolString(term2.expr);
    corremoveRelFromDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2));

    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corClearRelSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if ((iterator.arg_list.count() != 2) && (iterator.arg_list.count() != 1)) {
        CORAL_error(COR_BAD_ARITY, "1 or 2 arguments to rel_open needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a func-arg", "ClearRelSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    if (iterator.arg_list.count() == 1) {
        corclearRelInDB(term1.expr.FuncArgP(), corCurDB->name);
        return(iterator.bindenv);
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term2);
    cordeep_dereference_term (term2);
    if (term2.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "ClearRelSolver");
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    char *actual_string2 = corSymbolString(term2.expr);
    corclearRelInDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2));

    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* corCopyRelSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if ((iterator.arg_list.count() != 2) && (iterator.arg_list.count() != 3)) {
        CORAL_error(COR_BAD_ARITY, "2 or 3 arguments to rel_add needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a relation template",
        "CopyRelSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term2);
    cordeep_dereference_term (term2);
    if (term2.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "CopyRelSolver");
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    char *actual_string2 = corSymbolString(term2.expr);

    if (iterator.arg_list.count() == 2) {
        corcopyToDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2),
            term1.expr.FuncArgP());
        return(iterator.bindenv);
    }

    corTerm term3 (iterator.arg_list[2], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term3);
    cordeep_dereference_term (term3);
    if (term3.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a relation template",
            "CopyRelSolver");
        term3.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    corcopyToDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2),
        term3.expr.FuncArgP());

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv *corAddRelSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    if ((iterator.arg_list.count() != 2) && (iterator.arg_list.count() != 3)) {
        CORAL_error(COR_BAD_ARITY, "2 or 3 arguments to rel_add needed", NULL);
        FAIL_RETURN
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term1);
    cordeep_dereference_term (term1);
    if (term1.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a relation template",
            "AddRelSolver");
        term1.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    corTerm term2 (iterator.arg_list[1], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term2);
    cordeep_dereference_term (term2);
    if (term2.expr.kindof() != COR_SYMBOL_CONST) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a symbol", "AddRelSolver");
        term2.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    char *actual_string2 = corSymbolString(term2.expr);

    if (iterator.arg_list.count() == 2) {
        coraddToDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2),
            term1.expr.FuncArgP());
        return(iterator.bindenv);
    }

    corTerm term3 (iterator.arg_list[2], iterator.bindenv);
    //FULL_DEREFERENCE_TERM (term3);
    cordeep_dereference_term (term3);
    if (term3.expr.kindof() != COR_FUNCTOR) {
        CORAL_error(COR_BAD_ARG_TYPE, "should be a relation template",
            "CloseRelSolver");
        term3.printon(corexEnv->error_file);
        fprintf(corexEnv->error_file,"\n");
        FAIL_RETURN
    }

    coraddToDB(term1.expr.FuncArgP(), corEnterSymbol(actual_string2),
        term3.expr.FuncArgP());

    return(iterator.bindenv);        
}

#undef FAIL_RETURN

//**************************************************************************
// This is an example of how to 'macro' a number of builtins under
// a single name -- PRAVEEN. More work needs to be done on this
//**************************************************************************

corBindEnv *corListAllSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    char buf[BUFF_MAX];

    sprintf(buf, "list_dbs. list_rels.");
    corparseStack->handle_new_buff(buf,"buffer");
    corparseStack->parse(corexEnv->error_file,1,"Coral Internal Error.\n");

    return(iterator.bindenv);
}

//**************************************************************************
#define FAIL_RETURN {iterator.set_no_match(); return(iterator.bindenv);}

corBindEnv* corClearDBSolver(corBuiltinRelation&, corTupleIterator& iterator)
{
    char buf[BUFF_MAX];
    Name name;

    if (iterator.arg_list.count() > 1) {
        CORAL_error(COR_BAD_ARITY, "1 argument to ws_clear needed", NULL);
        FAIL_RETURN
    }

    if (iterator.arg_list.count() == 0)
        name = corCurDB->name;
    else {
        corTerm term1 (iterator.arg_list[0], iterator.bindenv);
        //FULL_DEREFERENCE_TERM (term1);
        cordeep_dereference_term (term1);
        if (term1.expr.kindof() != COR_SYMBOL_CONST ) {
            CORAL_error(COR_BAD_ARG_TYPE, NULL, "ClearDBSolver");
            term1.printon(corexEnv->error_file);
            fprintf(corexEnv->error_file,"\n");
            FAIL_RETURN
        }
        name = term1.expr.SymbolP();
    }

    sprintf(buf, "%s%s %s%s %s%s%s%s %s%s%s%s",
    corSymbolString(corCreateDBSymbol),"(\"__tmp\").\n",
    corSymbolString(corCurrentDBSymbol),"(\"__tmp\").\n",
    corSymbolString(corDestroyDBSymbol), "(", corSymbolString(name) , ").\n",
    corSymbolString(corRenameDBSymbol),"(\"__tmp\",", corSymbolString(name), ").\n");

    corparseStack->handle_new_buff(buf,"buffer");
    corparseStack->parse(corexEnv->error_file,1,"Coral Internal Error.\n");

    return(iterator.bindenv);
}

#undef FAIL_RETURN

//**************************************************************************
