//**************************************************************************
//    Solvers.h
//**************************************************************************

#ifndef _Solvers_H_
#define _Solvers_H_

//**************************************************************************

// this class is used for  StartTimeSolver(), EndTimeSolver,CputimeSolver
// because getrusage is not available on HP.

#include <unistd.h>
#include <sys/times.h>

#include "CoralError.h"

class corCcoralTimer {
  struct tms start, end, coral_start; // coral_start does not get reset!
  clock_t wall_start;		      
  int clktck ;
 public:
  double wall, sys, usr, coral_usr; // coral_usr is the accumulative usr time
  corCcoralTimer(){
    if ((clktck = sysconf(_SC_CLK_TCK)) < 0){
      CORAL_error(COR_FATAL_ERROR, "","Coral timer: can't call sysconf");
      clktck = 100;
     }
    times(&coral_start);
    reset();
   };
  ~corCcoralTimer(){};
  reset(){
    wall_start = times(&start);
   };
  look(){
    wall = (times(&end) - wall_start) / double (clktck) ;
    sys = (end.tms_stime - start.tms_stime + end.tms_cstime 
	   - start.tms_cstime) / double (clktck) ;
    usr = (end.tms_utime - start.tms_utime + end.tms_cutime 
	   - start.tms_cutime) / double (clktck) ; 
    coral_usr = (end.tms_utime - coral_start.tms_utime + end.tms_cutime 
	   - coral_start.tms_cutime) / double (clktck) ; 
   };
 };

class corBuiltinRelation;
class corTupleIterator;
class corBindEnv;
class corTuple;

//**************************************************************************

corBindEnv *corUnivSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corCallSolver(corTupleIterator& iterator);

corBindEnv *corMemberSolver(corTupleIterator& iterator);
corBindEnv *corMemTupleSolver(corTupleIterator& iterator);
corBindEnv *corMultiUnionSolver(corTupleIterator& iterator);
corBindEnv *corUnionSolver(corTupleIterator& iterator);
corBindEnv *corIntersectionSolver(corTupleIterator& iterator);
corBindEnv *corDiffSolver(corTupleIterator& iterator);
corBindEnv *corSubsetSolver(corTupleIterator& iterator);
corBindEnv *corMakeSetSolver(corTupleIterator& iterator);
corBindEnv *corUnionDSolver(corTupleIterator& iterator);
corBindEnv *corMultiUnionDSolver(corTupleIterator& iterator);
corBindEnv *corIntersectionDSolver(corTupleIterator& iterator);
corBindEnv *corDiffDSolver(corTupleIterator& iterator);
corBindEnv *corMakeSetDSolver(corTupleIterator& iterator);
corBindEnv *corCutSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corQuitSolver(corBuiltinRelation&, corTupleIterator& iterator);

corBindEnv *corResetCounterSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corCounterSolver(corBuiltinRelation& , corTupleIterator& iterator);

corBindEnv *corRelOptionsSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corEmptyPrintSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corDebugSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corHelpSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corConsultSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corTuplePrintSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corBindingsPrintSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corPrintSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corPrintfSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corStartTimeSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corEndTimeSolver(corBuiltinRelation& , corTupleIterator& iterator);
corTuple   *corCputimeSolver(corTuple *query);
corBindEnv *corShellSolver(corBuiltinRelation& , corTupleIterator& iterator);
corTuple   *corExtractSolver(corTuple *query);
corTuple   *corLengthSolver(corTuple *query);
corTuple   *corSubstringSolver(corTuple *query);
corBindEnv *corpipeOutSolver(corBuiltinRelation& , corTupleIterator& iterator);
corBindEnv *corwriteTableSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *correadTable(const char *infile,corTupleIterator& iterator);
corBindEnv *corpipeInSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *coraddIndexSolver1(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *coraddIndexSolver2(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *ArgTypeSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corHistorySolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corDisplayDefaultsSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corResetDefaultsSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corDeleteFactSolver(corBuiltinRelation&,corTupleIterator& iterator);
corTuple   *corPowSolver(corTuple* in);
corTuple   *corCreateSetSolver(corTuple *query);
corBindEnv *corAliasSolver(corBuiltinRelation&, corTupleIterator& iterator);
corTuple   *corAddElemDSolver(corTuple *query);
corTuple   *corAddElemSolver(corTuple *query);
corTuple   *corAbsSolver(corTuple *query);
corTuple   *corCeilSolver(corTuple *query);
corTuple   *corFloorSolver(corTuple *query);
corBindEnv *corInsertFactSolver(corBuiltinRelation&,corTupleIterator& iterator);


corBindEnv *corAnswerSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corLoginSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corLogoutSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corWhoSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corWhoamiSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corYieldSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corUserSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corKillSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corListSusSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corResumeSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corSyncSolver(corBuiltinRelation& rel, corTupleIterator& iterator);

corBindEnv *corBreakSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corClearBreakSolver(corBuiltinRelation&, corTupleIterator& iterator);
corBindEnv *corPrintModuleSolver(corBuiltinRelation&, corTupleIterator& iterator);

//**************************************************************************

#endif _Solvers_H_

