//**************************************************************************
//    main.C
//**************************************************************************

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "CoralInit.h"
#include "Globals.h"
#include "ParseStack.h"
#include "CoralError.h"
#include "Util.h"
#include "BuiltinUtil.h"
#include "ConnectArray.h"
#include "ClientProtocol.h"

#ifdef DO_RDB

#include <new.h>
#include <signal.h>

#include "Database.h"
#include "Execute.h"
#include "CoralBreak.h"
#include "InitBuiltins.h"
#include "DontCareArg.h"
#include "BuiltinSymbols.h"
#include "SMUtil.h"
#include "CRdbManager.h"

static void init_coral();

extern char coral_name[200];    // full pathname of the executing program

extern int corstartMemUsage;
extern struct rusage corstartRUsage;
extern struct rusage corSystemUsage;
extern int corbeginMem;

extern corArg* corTheParameterArg;

static void exceptionHandler(int recvd_signal);
static void brokenPipe(int recvd_signal);
static void freeStoreException();


#endif  // DO_RDB

static char coral_version[] = "Coral Version 1.2";

//**************************************************************************

void Command_Line_Error(char *str);
int  GetFilePointers(char **src_str, int max_files);
int  GetAnswerStyle(char **src_str, int max_args);
void Command_Line_Error(char *str);

//**************************************************************************

int main(int ac, char **av)
{
    int i;
    unsigned short use_port = 0;

      // Need to know if we are in server mode ASAP
    for (i=1; i < ac; i++){
        if (strcmp(av[i],"-s") == 0)
            coralConfig.server_mode = 1;

        if (strncmp(av[i],"-p",2) == 0){
            if (strlen(av[i]) > 2)
                sscanf(&(av[i][2]),"%hu", &use_port);
        }
    }

    if (coralConfig.server_mode){
          // Connect to the coral port
        int err;
        if (use_port)
            err = corclients->BindToPort(use_port);
        else
            err = corclients->BindToPort();

        // Could not bind to port (or something, specific error message
        // given by lower level.)
        if (err < 0)
            exit(-1);
    } else {

          // Start up a "Local" Connection (use stdin/stdout)
        corclients->LocalConnection();
    }

    for (i=1; i<ac && *av[i]=='-'; ++i) {

        switch (av[i][1]) {
        case 'a':
            // Answer style
            if (av[i][2] != '\0') Command_Line_Error(av[0]);
            i += GetAnswerStyle((char **)(av+i+1),(int)(ac - i -1));
            break;

        case 'c':
            // Read .coralrc mode ON : default
            corexEnv->C_read_coralrc_default = 1;
            break;

        case 'C':
            // Read .coralrc mode OFF
            corexEnv->C_read_coralrc_default = 0;
            break;

        case 'e':
            // Connect to EXODUS
            corexEnv->C_connect_exodus_default = 1;
            break;

        case 'E':
            // Don't connect to EXODUS : default
            corexEnv->C_connect_exodus_default = 0;
            break;

        case 'f':
        case 'F':
            // Input files
            if (av[i][2] != '\0') Command_Line_Error(av[0]);
            i += GetFilePointers((char **)(av+i+1), (int)(ac-i-1));
            break;

        case 'h':
        case 'H':
            // Print version and show usage 
            fprintf(stdout, "\nThis is %s\n\n", coral_version);
            Command_Line_Error(av[0]);
            break;

        case 'i':
            // Interactive mode ON : default
            corexEnv->C_interactive_mode_default = 1;
            break;

        case 'I':
            // Interactive mode OFF
            corexEnv->C_interactive_mode_default = 0;
            break;


        case 'm':
            // Create .P.M mode ON : default
            corexEnv->C_create_magic_default = 1;
            break;

        case 'M':
            // Create .P.M mode OFF
            corexEnv->C_create_magic_default = 0;
            break;

        case 'o':
            // Output file
            if ((av[i][2] != '\0')||(i == ac-1)) Command_Line_Error(av[0]);
            if (!set_print_file(av[++i])) {
                fprintf(stderr, "%s error: Cannot set print file to %s\n",
                    av[0], av[i]);
                exit(1);
            }
            break;

        case 'p':
            // Display the prompt : default
            corexEnv->C_print_prompt_default = 1;
            break;

        case 'P':
            // Don't display the prompt
            corexEnv->C_print_prompt_default = 0;
            break;

        case 'q':
            // Quiet mode ON
            corexEnv->C_quiet_mode_default = 1;
            break;

        case 'Q':
            // Quiet mode OFF : default
            corexEnv->C_quiet_mode_default = 0;
            break;

        case 's':
            // server mode ON : not default;
            coralConfig.server_mode = 1;
            break;

        case 'v':
            // Verbose answer display ON : default
            corexEnv->C_verbose_answer_default = 1;
            break;

        case 'V':
            // Verbose answer display OFF
            corexEnv->C_verbose_answer_default = 0;
            break;

        case 'w':
            // Insert mode ON : default
            corexEnv->C_insert_mode_default = 1;
            break;

        case 'W':
            // Insert mode OFF
            corexEnv->C_insert_mode_default = 0;
            break;

        default:
            Command_Line_Error(av[0]);
        }
    }

    if (i<ac) Command_Line_Error(av[0]);

        /* Clear the symbol table when the program exits */
    corclearSym.clear_at_end();

    corexEnv->C_exec_mode = COR_DOING_INTERPRET;
    init_coral(av[0]);


    if (coralConfig.server_mode){

          // Wait for a user
        corclients->NextConn();
    }


    if (!corexEnv->C_quiet_mode_default && !coralConfig.server_mode) {
           // Print out the version of coral
        fprintf(corexEnv->print_file, "\nThis is %s\n", coral_version);
          // Print preliminary message
        char name_buffer[1000] ;
        sprintf(name_buffer, "%s%s/init.msg", corexEnv->coral_path, HELP_PATH);
        corDisplayFile(name_buffer);
    }


    while (! corparseStack->is_empty()){

        corparseStack->parse(stdout,1,"syntax error -- skipping to next period\n");

        if (coralConfig.server_mode){

            if (!corparseStack->is_empty())
                corparseEnv->print_prompt();
            fflush(corexEnv->print_file);

            if (corcurConn && corparseStack->is_empty()){

                  // Close the socket and delete the connection
                corclients->DeleteConnection(corcurConn->conn_id);
            }

              // Switch users
            corclients->NextConn();
        }
    }
    
    exit_coral();
#ifdef DO_RDB
    delete pCRdbManager;
#endif  // DO_RDB

    if (!corexEnv->C_quiet_mode_default) {
        fprintf(corexEnv->print_file,"\nHave a good day, %s !\n", cuserid(NULL));
    }

    return(0);
}

//**************************************************************************
// Prints error message about command line error

void Command_Line_Error(char * str)
{
    fprintf(stderr, 
      "Input syntax : %s [-o/O <output-file>][-f/F <input-file1> ...]\n", str);
    fprintf(stderr, "      -a answer_style   Specify answer print style (bindings, tuples or empty)\n");
    fprintf(stderr,
      "      -c                Enable reading of .coralrc startup script\n");
    fprintf(stderr,
      "      -C                Disable reading of .coralrc startup script\n");
    fprintf(stderr,"      -e                Connect to EXODUS\n");
    fprintf(stderr,"      -E                Do not connect to EXODUS\n");
    fprintf(stderr,"      -f <in_file>%s     Specify input files\n", "+");
    fprintf(stderr,"      -h                Show usage (this message)\n");
    fprintf(stderr,"      -i                Enable interactive mode\n");
    fprintf(stderr,"      -I                Disable interactive mode\n");
    fprintf(stderr,"      -m                Enable creation of.P.M files\n");
    fprintf(stderr,"      -M                Disable creation of.P.M files\n");
    fprintf(stderr,"      -o output_file    Specify output file\n");
    fprintf(stderr,"      -p                Enable prompt printing\n");
    fprintf(stderr,"      -P                Disable prompt printing\n");
    fprintf(stderr,"      -q                Enable quiet mode\n");
    fprintf(stderr,"      -Q                Disable quiet mode\n");
    fprintf(stderr,"      -s                Enable server mode\n");
    fprintf(stderr,"      -v                Enable verbose answer display\n");
    fprintf(stderr,"      -V                Disable verbose answer display\n");
    fprintf(stderr,"      -w                Enable insert mode\n");
    fprintf(stderr,"      -W                Disable insert mode\n");

    fprintf(stderr,"\nSpecial when in server mode:\n");
    fprintf(stderr,"\n      -pNNNN          NNNN is the port to use\n");
    exit(1);
}

//**************************************************************************
/*
 * Reads the filenames specified in src_strin, opens the files, and saves
 * the respective parsers onto the parseStack.  Also deletes the .M files
 * corresponding to the opened files.
 */

int GetFilePointers(char ** src_str, int max_files)
{
    int count = 0;
    FILE *tmp_fp;
    char temp_str[200];
    
    if (coralConfig.server_mode){
        fputs("File loading not supported in server mode\n",stderr);
        exit(0);
    }

      // Remove stdin from stack
    corparseStack->pop();

    while ((count < max_files) && (*src_str[count] != '-')) {
        if (tmp_fp = fopen(src_str[count],"r")) {
            corparseStack->put_on_bottom(tmp_fp,0,src_str[count]);
            sprintf(temp_str, "%s.M", src_str[count]);
            unlink(temp_str);
        } else {
            CORAL_error(COR_FILE_NOT_FOUND, src_str[count], "GetFileList",
                __FILE__, __LINE__);
        }
        count++;
    }

      // Put stdin back on the bottom
    corparseStack->put_on_bottom(stdin,0,src_str[count]);
    corparseStack->reset_position();

    return(count);
}

//**************************************************************************
// Reads the answer print style from the command line

int GetAnswerStyle(char** src_str, int max_args)
{
    if (max_args <= 0) return(0);

    if (!(strcmp(src_str[0],"tuple") && strcmp(src_str[0],"tuples")))
        corexEnv->C_answer_style_default = COR_TUPLE_ANSWER_PRINT;
    else if (!(strcmp(src_str[0],"binding") && strcmp(src_str[0],"bindings")))
        corexEnv->C_answer_style_default = COR_BINDING_ANSWER_PRINT;
    else if (!strcmp(src_str[0],"empty"))
        corexEnv->C_answer_style_default = COR_EMPTY_ANSWER_PRINT;
    else
        Command_Line_Error(src_str[0]);

    return(1);
}

//**************************************************************************
//    Initializes CORAL.  Called from main routine.

#ifdef DO_RDB

static void init_coral(char *pgm)
{
#    ifndef DEBUG
        // Catch signals that might otherwise cause termination
        for (int i = SIGINT; i < SIGPIPE; i++) {
            if (i != SIGKILL) {
#                ifdef __GNUC__
                    signal(i, (exceptionHandler)) ;
#                else
                    signal(i, (COR_SIGNAL_HANDLER)(exceptionHandler)) ;
#                endif
            }
        }

          // Catch broken pipes (An Explain that quit)
#       ifdef __GNUC__
            signal(SIGPIPE, (brokenPipe));
#       else
            signal(SIGPIPE, (COR_SIGNAL_HANDLER)(brokenPipe));
#       endif

#    else
        // for debugging purposes, catch only SIGINT (CTRL-C)
#        ifdef __GNUC__
            signal(SIGINT, (exceptionHandler)) ;
#        else
            signal(SIGINT, (COR_SIGNAL_HANDLER)(exceptionHandler)) ;
#        endif
#    endif


    sprintf(coral_pid, "%d", getpid());

    // If neither LocalConnection() nor BindToPort() has been called,
    // set up a local connection with an empty stack.
    if (! corclients->serv_sock)
        corclients->LocalConnection(0);

    // Provide exception handler for new()
    set_new_handler(freeStoreException);

    // create the CurrModuleInfo structure
    corCurrModuleInfo = new corModuleInfo();

    // create the functor table
    corcreateFunctorTable();

    // init builtin symbols
    corinit_builtin_symbols();

    strcpy(coral_name, (pgm ? pgm : "coral"));

    if ((corexEnv->C_exec_mode == COR_DOING_INTERPRET) ||
        (corexEnv->C_exec_mode == COR_DOING_IMPERATIVE))
    {
        // initialize global stats for CPU and memory usage
#        ifndef UNIX_HP
            getrusage(RUSAGE_SELF, &corstartRUsage);
#        endif
        corstartMemUsage = (int)sbrk(0);

#        ifdef WITH_PERSISTENCE
            if (corexEnv->C_connect_exodus_default) {
                // Create exodus workspace
                corCurDB = &corExodusDB;        
                corCurDB->RelationTable = corAllocTab(COR_RELATION_TABLE_INCR, NULL);    
                // Add exodus database to the Database Table
                if (!coraddDBStruct(corCurDB)) {
                    CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add exodus DB",
                        "init-coral", __FILE__, __LINE__);
                }
                else 
                    sm_init();
            }
#        endif

#       ifdef DO_RDB
            pCRdbManager = new corCRdbManager();
#       endif

        // Initialize TheParameterArg used for C++ host variables
        corTheParameterArg = &corTheDontCareArg;

        // Remove the stdin.M file
        unlink("stdin.M");

        // Initialize builtin database.
        corBuiltinDB.name = corEnterSymbol("builtin_ws");
        corinitBuiltins();
        corCurDB = &corBuiltinDB;

        // Add builtin database to the Database Table
        if (!coraddDBStruct(corCurDB)) {
            CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add builtin DB",
                "init-coral", __FILE__, __LINE__);
            exit(1);
        }

        // Create default database
        corDatabaseStruct *db = corcreateDB(corEnterSymbol(COR_DEFAULT_DB));
        if (!db) {
            CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add default DB",
                "init-coral", __FILE__, __LINE__);
            exit(1);
        }
        corCurDB = db ;

        if (corexEnv->C_read_coralrc_default) {
            // Read .coralrc if it exists, else read input from $(HOME)/.coralrc
            if (!coralConfig.server_mode &&
                 !corconsultFile(".coralrc", ".coralrc")) {
                char *home_path = getenv("HOME") ;
                char *tmp_name = new char[strlen(home_path) + 15] ;
                strcpy(tmp_name, home_path);
                strcat(tmp_name, "/.coralrc");
                corconsultFile(tmp_name, tmp_name);
                delete [] tmp_name;
            }
        }
    }
    else {
        corexEnv->C_do_history_default = 0;
    }

#    ifndef UNIX_HP
    getrusage(RUSAGE_SELF, &corSystemUsage);
#    endif
    corbeginMem = (int)sbrk(0);
}

//**************************************************************************
//    Signal handler

static void exceptionHandler(int recvd_signal)
{
    CORAL_error(COR_SIGNAL_RECEIVED,NULL,NULL,__FILE__,__LINE__);
    fprintf(corexEnv->error_file, "Received signal no: %d\n", recvd_signal);

    //    Code for recovering from abnormal situation
    if (recvd_signal != SIGINT) {
        fprintf(corexEnv->error_file, "Cannot recover from received signal\n");
        fprintf(corexEnv->error_file, "Initiating shutdown procedures\n");
#        ifdef WITH_PERSISTENCE
            abortTrans();
#        endif
        exit_coral();
#ifdef  DO_RDB
        delete pCRdbManager;
#endif  // DO_RDB 
        exit(1);
    }

    if (corexEnv->C_interrupt_raised == 1) {
        fprintf(corexEnv->error_file, "Received second interrupt\n");
        fprintf(corexEnv->error_file, "Initiating shutdown procedures\n");
#        ifdef WITH_PERSISTENCE
            abortTrans();
#        endif
        exit_coral();
#ifdef  DO_RDB
        delete pCRdbManager;
#endif  // DO_RDB 
        exit(1);
    }

    corexEnv->C_interrupt_raised = 1;
}

//**************************************************************************
//    Signal handler for broken pipe (caused by Explain that disconnects)

static void brokenPipe(int recvd_signal)
{
    // Assume it was the current users explain (pretty safe assumption)

      // Turn off explain and reset associated structures
    corcurUser->corexEnv.explain_fp = (FILE *) NULL;

    corcurUser->corexEnv.GlobalRelationOptions &= ~REL_EXPLAIN;

      // Notify user
    if (corcurUser->corexEnv.error_file)
        fputs("Explain disconnected - turning explain off.\n",
                            corexEnv->error_file);

    if (corcurUser->explain_sock){
        delete corcurUser->explain_sock;
        corcurUser->explain_sock = (CoralSock *) NULL;
    }
}

//**************************************************************************
//    Exception handler for the new() system call

static void freeStoreException()
{
    CORAL_error(COR_NO_MEMORY,NULL,"freeStoreException",__FILE__,__LINE__);
    fflush(corexEnv->print_file);
    fflush(corexEnv->error_file);
#ifdef  DO_RDB
        delete pCRdbManager;
#endif  // DO_RDB 
    exit(1) ;
}

#endif /* DO_RDB */

