/*****************************************************************************/
//    GenericIndex.C
/*****************************************************************************/

#include "Globals.h"
#include "GenericIndex.h"

#include "Interface.h"

/*****************************************************************************/
int corGenericIndex::matches(corGenericIndex *ind)
{
    if (argset) {
        if (!ind->argset) return 0;
        if (*argset == *(ind->argset))
            return 1;
        return 0;
    }

    return 1;
}

/*****************************************************************************/
void corGenericIndex::delete_all_tuples()
{
    fprintf(corexEnv->error_file, "delete_all_tuples: not implemented yet\n");
}

/*****************************************************************************/
/* 
    This takes a pattern form index and returns an argument form
    index, if possible; else it returns NULL.

        (1) if there is a function symbol in pat with an embedded variable 
            that does not occur in v_nm, return NULL
        (2) if there is a repeated variable in pat which does not occur 
            in v_nm, return NULL
*/

char *corsimplify_index(corArgList *pat, int n_var, corArgList *v_nm)
{

    corStackMark stackmark;
    corArrayBindEnv pat_env(n_var);
    corTerm dummy_term(corZeroCArg, NULL);

#    ifdef DO_TRACE
        if (corexEnv->dbg_indexing) {
            fprintf(corexEnv->trace_file, "simplify_index: (");
            pat->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file, "), bound vars = (");
            v_nm->printon(corexEnv->trace_file);
            fprintf(corexEnv->trace_file,")\n");
        }
#    endif

    // First bind each position of pat_env occurring in v_nm to dummy_term
    for (int i=0; i < v_nm->count(); i++) {
        pat_env.bind((*v_nm)[i].VarArgP()->var, dummy_term);
    }

    // Now check if each argument of pat is either a variable or a constant
    //     in pat_env

    char *return_val = new char[pat->count()+1];
    for (int j=0; j < pat->count(); j++) {
        if ((*pat)[j].hash(&pat_env) == VarHashValue) {
            if ((*pat)[j].kindof() != COR_VARIABLE){
                // embedded unbound variable
                delete return_val;
                stackmark.pop_to();
                return NULL;
            }
            else {
                // free variable
                return_val[j] = 'f';
            }
        }
        else {
            // the argument is ground, given the bound variables
            return_val[j] = 'b';
        }
    }
    return_val[j] = '\0';

    // Now check for repeated unbound variables at top-level in pat.

    for (int j1=0; j1 < pat->count(); j1++){
        if (((*pat)[j1].kindof() == COR_VARIABLE) && (return_val[j1] == 'f')) {
            // check if this is a repeated variable
            for (int j2=j1+1; j2 < pat->count(); j2++) {
                if (((*pat)[j2].kindof() == COR_VARIABLE) &&
                    ((*pat)[j1].VarArgP()->var == (*pat)[j2].VarArgP()->var))
                {
                    // found a repeated variable
                    delete return_val;
                    stackmark.pop_to();
                    return NULL;
                }
            }
        }
    }

#    ifdef DO_TRACE
        if (corexEnv->dbg_indexing) {
            fprintf(corexEnv->trace_file,
                "simplify_index: simplified to %s\n", return_val);
        }
#    endif

    stackmark.pop_to();
    return return_val;

}

/*****************************************************************************/

