//**************************************************************************
//    ParseStack.h
//
//    Shaun Flisakowski - Oct 1994
//**************************************************************************

#ifndef _ParseStack_H_
#define _ParseStack_H_

#include "ParserEnv.h"

//**************************************************************************
//
// This structure is a stack of parser structures to be used by the
// parser. The topmost ParserEnv is the current one in use.
// When it reaches EOF, it is popped, and yyparse() terminates. Since
// yyparse() may now be called recursively, this does not necessarily
// mean the game is over. 
// To start parsing from a new file pointer or buffer, push it onto 
// the stack.
//
//**************************************************************************

#define    YYBUFF_SIZE    4096

class corParseStack {

private:
    corParserEnv *top;
    corParserEnv *bottom;

public:

    int  parse(FILE *err_fp, int parse_cnt = 1,
                 char *err_mess = "Parse Error.\n");
    void handle_new_file(FILE *fp, char *fname );
    void handle_new_buff(char *buff, char *bname );
    int  get_next_file();
    void abort_file();
    void show_stack();
    
    inline char *cur_file_name() { return(top->filename); }

    void reset_position();
    void push(FILE *fp1, int line_num = 1, char *fname = NULL);
    void push(char *parsebuff, int line_num = 1, char *bname = NULL );
    void put_on_bottom(FILE *fp1, int line_num = 1, char *fname = NULL);

    void pop();

    inline corParserEnv *top_of_stack() { return(top); }

    inline char *top_file_name() {
        if ((top) && (top->filename))
            return(top->filename);
        else
            return(NULL);
    }

    inline FILE *top_file() { return(top->fp); }
    inline int top_file_line() { return(top->cur_line_num);}

    inline void set_top_file_line(int i) {
        if (top) top->cur_line_num=i;
    }

    inline int is_empty() { return(top == NULL); }

    corParseStack() { top = NULL; bottom = NULL; }
   ~corParseStack();
};

//**************************************************************************

#endif _ParseStack_H_

