//**************************************************************************
//    ParserStruct.h
//**************************************************************************

#ifndef _ParserStruct_H_
#define _ParserStruct_H_

#include "Rule.h"
#include "ModuleInfo.h"
#include "Table.h"
#include "RuleInfoArray.h"
#include "SemiNaive.h"
#include "PredAnnotations.h"
#include "ParseUtils.h"
#include "Array.h"
#include "Collection.h"
#include "SymbolArg.h"

//**************************************************************************

class corParserStruct;
typedef void (* ExecFunction) (corParserStruct&);

//**************************************************************************

class corParserStruct {
public:
    ExecFunction func;
    int func_setup;
    corClause *cur_rule;
    int label_counter;

    corSemiNaive *snrule;
    corTable *table;
    corRuleInfoArray *rInfo_arr;

    corModuleInfo CurModule;
    int nested_index;

    // The type 'arg_link' is used to temporarily chain together the
    // arguments of a functor or predicate.
    // The variable 'cur_arg_chain' contains a list of the arguments of
    // the functor or predicate that we're currently scanning.
    // Surrounding functors or predicates are pushed using the
    // 'pushed' link.

    struct corparameter_chain *parameters;
    struct corarg_chain *cur_arg_chain;
    corset_stack *cur_set;
    struct corarray rule_var_names;
    corCArgDynArray sip_var_names;
    struct corarray rhspreds;
    corClause *rule_list;
    struct corcollection rules[1];
    struct corPredicate *Predicates;
    struct corSymbolTable *PredicateTable;
    corPredicate *AllPredicates;
    corPredAnnotations *make_index_annotations;
    corPredAnnotations *agg_sel_annotations;
    corPredAnnotations *prioritize_annotations;
    corPredAnnotations *diskrel_annotations;
    corPredAnnotations *AllowedAdornList;
    corPredAnnotations *DisallowedAdornList;

    // WARNING: In case of conflicts between AllowedAdornList and 
    //          DisallowedAdornList, the DisallowedAdornList is
    //          given priority, and the AllowedAdornList is over-ridden.

    corPredAnnotations *MultisetAnnotations;
    int lhs_flag;

    void init();

    void assign_func(ExecFunction exec_func) {
        func_setup = 1;
        func = exec_func;
    }

    void execute() { if (func_setup) (*func)(*this); }

    // to be called by the parser at the end of each input parserStruct to
    // release unnecessary memory
    void reset(int flag = 0);

    // This clears the memory but does not reinitialze the parserStruct
    // (used by reset() and the destructor)
    void clear(int flag = 0);

    corParserStruct() {
        parameters = NULL;
        init();
    };
   ~corParserStruct();
};

//**************************************************************************

corPredicate *corFindPredicate(corParserStruct& corparserStruct, Name name, int arity);
corPredicate *corLookupPredicate(corParserStruct& corparserStruct, Name name, int arity);

//**************************************************************************

#endif _ParserStruct_H_

