/*****************************************************************************/
//    PipelinedExecInfo.C
/*****************************************************************************/

#include <string.h>

#include "PipelinedExecInfo.h"

#include "QFPipelinedModuleData.h"
#include "TupleIterator.h"

/*****************************************************************************/

#define NULL_INT_VAL -2

int corC_PipelinedConstructs = 0;
int corC_PipelinedDestructs = 0;

/*****************************************************************************/
corPipelinedExecInfo::corPipelinedExecInfo(corQFPipelinedModuleData *md, int pred_offset)
{
    corRelation *rel;
    int offset;

    // Get the first rule defining this relation
    if (!(curRule = (md->pred_ptrs[pred_offset])->clauses)) return;

    // Extract its RuleInfo 
    rInfo = (md->ruleArray)->rInfoPtrs[curRule->rule_number];

    // Table is the same as that in the QFPipelinedModuleData
    F = md->F;
    indices = md->indices[curRule->rule_number];
    int numVars = md->maxNumVars[pred_offset];
    env = new corArrayBindEnv(numVars);
    int numLits = md->maxNumLiterals[pred_offset];
    tuple_its = new corTupleIterator* [numLits];
    //bzero((char *)tuple_its, numLits*sizeof(TupleIterator *) );
    memset((char *)tuple_its, 0, numLits*sizeof(corTupleIterator *) );
    tuples = new corTuple* [numLits];
    stack_marks = new corStackMark [numLits];

    firstFailureBacktrack = md->backtrackInfo.
        firstFailureBacktrackArrays[curRule->rule_number];
    nextFailureBacktrack = md->backtrackInfo.
        nextFailureBacktrackArrays[curRule->rule_number];
    successBacktrack = md->backtrackInfo.
        successBacktrackPoints[curRule->rule_number];

#    ifdef DEBUG
        corC_PipelinedConstructs++;
#    endif
    
    for (index = 0; index < rInfo->num_literals; index++)
    {
        //Check if predicate should be negated, signalled by an offset <= -2.
        if (NEGATED(index))
            offset = (-2) - rInfo->relations[index];
        else
            offset = rInfo->relations[index];
    
        rel = F->rels[offset].relPtr;
        tuple_its[index] = new corTupleIterator (rel, rInfo->arg_list[index],
            env, NULL, NULL, &(indices[index]));
    }
    index = current = backtrack_index = 0;
    direction = FORWARD;
    cut_crossed = 0;
}

/*****************************************************************************/
int corPipelinedExecInfo::init()
{
    return(0);
}

/*****************************************************************************/
corPipelinedExecInfo::~corPipelinedExecInfo()
{
#    ifdef DEBUG
        corC_PipelinedDestructs++;
#    endif

    if (!curRule) return;

    for (int i=0; i< rInfo->num_literals; i++) {
        if (tuple_its[i]) {
            tuple_its[i]->release();
            delete tuple_its[i];
        }
    }
    delete    tuple_its;
    delete tuples;
    delete stack_marks;
    indices = NULL;
    curRule = NULL;
    rInfo = NULL;
    F = NULL;
}

/*****************************************************************************/
/* 
    Go to the next rule that defines this predicate. 
    Sets up the info in PipelinedExecInfo

    Returns 1 on success, 0 on failure
*/

int corPipelinedExecInfo::NextRule(corQFPipelinedModuleData *md)
{
    corRelation *rel;
    int offset;

    // if we are already at the last rule defining the predicate
    if (!curRule->pnext) {
        return 0;
    }

    // use the structures allocated earlier
    // after resetting their values
    // The trail has been used already to reset values in the bindenvs
    for (int i=0; i< rInfo->num_literals; i++) {
        if (tuple_its[i]) {
            tuple_its[i]->release();
            delete tuple_its[i];
        }
        // There might be a memory leak here
        tuple_its[i] = NULL;
        tuples[i] = NULL;
    }
    
    curRule = curRule->pnext;

    // Extract its RuleInfo 
    rInfo = (md->ruleArray)->rInfoPtrs[curRule->rule_number];
    indices = md->indices[curRule->rule_number];

    // Initialize all other fields
    firstFailureBacktrack = md->backtrackInfo.
        firstFailureBacktrackArrays[curRule->rule_number];
    nextFailureBacktrack = md->backtrackInfo.
        nextFailureBacktrackArrays[curRule->rule_number];
    successBacktrack = md->backtrackInfo.
        successBacktrackPoints[curRule->rule_number];

    for (index = 0; index < rInfo->num_literals; index++) {
        //Check if predicate should be negated, signalled by an offset <= -2.
        if (NEGATED(index))
            offset = (-2) - rInfo->relations[index];
        else
            offset = rInfo->relations[index];
    
        rel = F->rels[offset].relPtr ;
        tuple_its[index] = new corTupleIterator (rel, rInfo->arg_list[index], env,
            NULL, NULL, &(indices[index]));
    }
    //backtrack_index = NULL_INT_VAL;
    //new_iterator = 0;
    offset = current = backtrack_index = 0;
    direction = FORWARD;
    return 1;
}

/*****************************************************************************/

