// *************************************************************************
//    Rule.h
// *************************************************************************

#ifndef _Rule_H_
#define _Rule_H_

#include "Link.h"
#include "SymbolArg.h"
#include "ClauseName.h"
#include "Annotations.h"
//#include "Literal.h"

// *************************************************************************

enum rule_type {
    COR_NORMALRULE,
    COR_MAGICRULE
};

// *************************************************************************

class corLiteral;

// *************************************************************************
// Rule structure created by the parser

class corClause {
public:
    Link link;                // All rules are chained together.
                              // Rules with the same head->pred are 
                              // grouped together.
    corClause* pnext;            // next, linked by predicate

    int num_var_names;
    Name* var_names;          // List of all variable names used in the rule.
    int old_num_var_names;    // Used in Magic rewriting 

    corLiteral* head;
    corLiteral** head_deletes;    // Array of delete literals in head
    int    head_deletes_count;

    int    num_preds;
    corLiteral** preds;           // List of rhs predicates
    struct corsips* sips;
    struct corRuleAnnotes* annotations; 

    int    last_part;          // last pred participating in sip
    int    line_number;        // source line of definition
    int    rule_number;        // unique rule number
    rule_type type;            // COR_NORMAL/COR_MAGIC 
    corClause  *source_rule;      // For COR_MAGIC rule types only 
    corLiteral *source_pred;      // For COR_MAGIC rule types only
    unsigned int recursive : 1; 
    unsigned int seen : 1;     // flag used during scc analysis

    struct corClauseName* clause_names;

    int    factor_status;        // for factoring code only
    corLiteral* factor_pred;    // the rhs pred which is factorable
    int     ext_magic;     // Added by Ken: external magic flag
    int     sn_rule_label_counter;  // added by Ken
    int     supp;  // used in rewriting for nested_scc_eval ONLY.
                   // 0 for ordinary magic, 1 for sup_magic

    corClause();
   ~corClause();
};

// *************************************************************************

#define FOR_EACH_CLAUSE_OF(rule, pred) \
  for (rule= (pred)->clauses; rule!=*(pred)->clause_tail; rule= (rule)->pnext)

void coralloc_preds(corClause *rule, int count);
corClause *corcopy_rule(corClause *prototype);

// *************************************************************************

#endif _Rule_H_

