//**************************************************************************
//Suspension.C
//**************************************************************************

#include <stdio.h>
#include "ArgList.h"
#include "TupleIterator.h"
#include "Tuple.h"
#include "Term.h"
#include "BindEnv.h"
#include "Execute.h"
#include "DerivedRelation.h"
#include "BuiltinUtil.h"

#include "Globals.h"

#include "Suspend.h"
#include "Suspension.h"

#include "ClientProtocol.h"

//**************************************************************************

corsuspensionClass corsuspension;
extern int corcurrentDB(Name name);

//**************************************************************************
corsuspensionClass::corsuspensionClass()
{

  set =-1;
  byte_count = corserver_page_length;
  for(int i=0; i<cormax_suspensions; i++) 
    context[i]=NULL;
}

//**************************************************************************
corsuspensionClass::~corsuspensionClass()
{

  for(int i=0; i<cormax_suspensions; i++) 
    if (context[i]){
        corSuspendContext *cid = context[i];
        delete cid->semi_naive;
        delete cid->table;
        delete context[i];
    }
}

//**************************************************************************
//default: list only the current user's. list(-2) for all users on this
// connection
//**************************************************************************
int corsuspensionClass::list(int id)
{
    int found =0;
    int cid = corcurConn->conn_id;

    for(int i=0; i<cormax_suspensions; i++) 
        if (context[i])
            if ((cid == context[i]->conn_id)
                        && ((id ==-2) || (id == context[i]->user_id))) {
                fprintf(corexEnv->print_file,
                      "ID=%d WS=%s SN=%p TB=%p ST=%p UID=%d ANS=%d\n",
                      i, corSymbolString(context[i]->name),
                      context[i]->semi_naive, context[i]->table,
                      context[i]->trail, context[i]->user_id,
                      context[i]->answer);
                found =1;
            }
    return(found);
}

//**************************************************************************
int corsuspensionClass::avail_id(void)
{
  for (int i=0; i<cormax_suspensions; i++) 
    if (context[i]==NULL) return(i);
  return(-1);            // no more id
}


//**************************************************************************
int corsuspensionClass::suspend(corSuspendContext* s)
{
  int id;
  if (( id=avail_id()) != -1 ) {
    context[id]=s;
    fprintf(corexEnv->print_file,"%s%d\n",CP_R_ID,id);
    //fprintf(exEnv->print_file,
            // "Bytes Sent = %d.", server_page_length - byte_count);
    set =-1;
    byte_count = corserver_page_length;
    return(0); //successful
  } else {
    fprintf(corexEnv->error_file,
        "Maximum number of allowed suspensions has reached.\n");
    return(-1);
  }
}

//**************************************************************************
int corsuspensionClass::resume(int id)
{
  corStackMark stack_mark;

  if ((id>=cormax_suspensions)||(context[id]==NULL)) {
    fprintf(corexEnv->error_file, 
        "%sInvalid id, can't resume suspended query.\n",CP_CERR);
    return(-1);
  }else {
#ifdef DEBUG
    fprintf(corexEnv->print_file,"Suspension %d is resuming ",id);
    fprintf(corexEnv->print_file,"in %s(S%pT%p) Stack=%p:\n",
            corSymbolString(context[id]->name), 
            context[id]->semi_naive, context[id]->table, context[id]->trail);
#endif

      // Check to make sure it belongs to this person
    if ((corcurConn && (corcurConn->conn_id != context[id]->conn_id)) ||
            (corcurUser->user_id != context[id]->user_id)) {
      fprintf(corexEnv->error_file, 
          "%sInvalid id, can't resume suspended query.\n",CP_CERR);
      return(-1);
    }
    
    if (corcurrentDB(context[id]->name) == 0) 
      return -1;        // switch work space, change prompt
    strcpy(corparseEnv->prompt_name, corSymbolString(context[id]->name));
    strcat(corparseEnv->prompt_name, coralConfig.prompt);
    fprintf(corexEnv->print_file,"PRMT:%s\n",corparseEnv->prompt);
 
    delete corCurTrail;
    corCurTrail = context[id]->trail; // switch stack

    corSuspendContext *cid = context[id];
    context[id]=NULL;        // free id

    if (coralConfig.server_mode) {// automatic suspension;
      set = corconvert_ans_2_set(cid->answer);
    }

    int num_ans = (cid->semi_naive)->evaluate(cid->table, 
                          NULL, cid,1); // continue eval
    if (num_ans == 0){
        delete cid->semi_naive;
        delete cid->table;
        delete cid;
    }

    if (corexEnv->C_verbose_answer_default) 
      fprintf(corexEnv->print_file,"(Number of Answers = %d)\n",num_ans);
    stack_mark.pop_to();    // clear stack for next query
    
    return(0);
  }
}

//**************************************************************************
int corsuspensionClass::kill(int id)
{
  if ((id>=cormax_suspensions)||(context[id]==NULL)) {
    fprintf(corexEnv->error_file, 
        "%sInvalid id, can't kill suspended query.\n",CP_CERR);
    return(-1);
  }

    // Check to make sure it belongs to this person
  if ((corcurConn && (corcurConn->conn_id != context[id]->conn_id)) ||
          (corcurUser->user_id != context[id]->user_id)) {
    fprintf(corexEnv->error_file, 
        "%sInvalid id, can't kill suspended query.\n",CP_CERR);
    return(-1);
  }
                // ???? questions about allocation
  delete context[id]->trail;    // stack
  corSuspendContext *cid = context[id];
  delete cid->semi_naive;
  delete cid->table;
  context[id]=NULL;        // free id
  fprintf(corexEnv->print_file,"Suspended query %d killed.\n",id);
  return(0);
}

//**************************************************************************
// killall() -
// Kill all queries belonging to this user.
//**************************************************************************
int corsuspensionClass::killall(int conn_id, int user_id)
{
  int i;

  for (i=0; i < cormax_suspensions; i++){

      // Delete suspended queries belonging to this person
    if (context[i] && (context[i]->conn_id == conn_id)
                        && (context[i]->user_id == user_id)){

      delete context[i]->trail;
      corSuspendContext *cid = context[i];
      delete cid->semi_naive;
      delete cid->table;
      context[i]=NULL;
    }
  }

  return(0);
}

//**************************************************************************
// value for number of answers per query: 
//    0 - suspend according to page length.
//    n - suspend after n answers
// 
// value for set(a member in this class) before calling evaluate:
//    0 - suspend according to page length
//    n - suspend after N-1 answers. 
//**************************************************************************

int corconvert_ans_2_set(int ans)
{
  if (ans > 0) 
    return ans + 1;
  else
    return ans;
}
//**************************************************************************
