
//**************************************************************************
//    ConnectArray.h
//
//    The ConnectArray Class - Used to hold an array of connections.
//
//**************************************************************************

#ifndef _ConnectArray_H_
#define _ConnectArray_H_

#include "Connection.h"

class CoralSock;

//**************************************************************************

                                    // The maximum number of physical
                                    // connections allowed. 
#define  MAX_CONNECTIONS        64

//**************************************************************************

class corConnectArray {

protected:
  int            cur_conn;      // The current connection 

  int            num_conn;      // The current number of connections

                                // The maximum number of physical connections
                                // allowed (at once) by this ConnectArray
                                // This can be changed dynamically
                                // (maybe to control load)
  int            max_conn;      

                                // A spot "likely" to be free in the 
                                // connection array.
  int            nxt_conn;

public:  


  CoralSock     *serv_sock;     // The server socket
  CoralSock     *expl_sock;     // The server socket (for explain)

  corConnection    *connArray[MAX_CONNECTIONS];

  fd_set         select_msk;    // A mask to use for select

  corConnectArray();
 ~corConnectArray();

    // Return the current number of connections
  inline int NumConnections() { return(num_conn); }

    // Return the maximum number of connections
  inline int MaxConnections() { return(max_conn); }

    // Set the maximum number of connections (must be <= MAX_CONNECTIONS)
  inline int SetMaxConnections(int i) {
        max_conn = (i < MAX_CONNECTIONS) ? i : MAX_CONNECTIONS;
  }

    // Make a new connection
  int NewConnection();

    // Delete a connection
  void DeleteConnection(int conn_indx);

    // Establish a connection for Explain
  int ExplainConnection();

    // Set up this array as a server
  int BindToPort(unsigned short port = CORAL_PORT);

    // Make a "Local" connection (when you're not in server mode)
    // use_stdin indicates you would like stdin put on the stack
  int LocalConnection(int use_stdin = 1);

    // Get the next connection with data and set it up to run
  int NextConn();

    // Return the number of connections with data pending
  int NumWaiting();

    // Switch to this connection (switches to current user there)
  inline int  useConnection(int conn_id) { 
    return( (connArray[conn_id])->useConnection() );
  }

    // Return a pointer to the current connection
  inline corConnection *currentConn() {
      return(connArray[cur_conn]);
  }

    // List users for the connection specified, if conn is Null, list all
    // users on all connections
  int ListWho(FILE *fp, corConnection *conn = NULL);
};

//**************************************************************************

#endif  // _ConnectArray_H_
