// *************************************************************************
//    SupMagic.C
// *************************************************************************

#include <string.h>

#include "SupMagic.h"

#include "Rule.h"
#include "Interface.h"
#include "Globals.h"
#include "Unparse.h"
#include "CompileRule.h"
#include "BuiltinSymbols.h"
#include "Literal.h"
#include "ExportInfo.h"
#include "DontCareArg.h"

// *************************************************************************

#ifndef MAX_NAME_LENGTH
#define MAX_NAME_LENGTH (256+1)
#endif

// *************************************************************************

extern int opt_share_functors;

// *************************************************************************

static corLiteral* build_pred(corClause *adrule, corLiteral *pred, int argnum);
static corLiteral* build_goalid(corClause *adrule, int argnum, int usemagic);
static corVarLink *vars_from(corClause *adrule, int pred);
static corLiteral* build_supmagic(corClause *adrule, int rule_num, int pos,
    corVarLink *var_list, int next_der_lit);
static corLiteral* make_fail_literal();

// *************************************************************************

static int dont_create_sup0 = 1;

//    WARNING:: DANGEROUS FEATURE - 
//    If add_fail_literal is set, add a fail literal to magic rules - 
//    the goalid builtin predicate in the supplementary rule must be fixed 
//    to do the job of creating magic facts, and the magic rule is just 
//    a dummy rule.  Note that a patch in interp.C ensures that fail is 
//    always the first literal to be evaluated in all semi-naive rules.
#ifdef INSERT_GOALID
static int add_fail_literal = 1; 
#else
static int add_fail_literal = 0;
#endif

static int rule_num = 0;

// *************************************************************************
/*
    This function does an enhanced form of supplementary magic, where each
    rule can have several body literals, but only the first two literals
    can have derived predicates (i.e. predicates defined in the given module.  
    Semi-naive rewriting of such rules can safely permute these two derived
    literals if required, without causing any join order problems.
*/

void supmagic(struct corcollection *ad_rules, struct corcollection *magic_rules)
{
    rule_num = 0;
    FOR_EACH_RULE(adrule, ad_rules) {
        //    Construct the supplementary magic rules.
            if (0 && corparserStruct.CurModule.NestedSccEval && !adrule->supp)
          continue; // added by Ken

        do_supp_magic_on_rule(adrule, magic_rules);
    } END_EACH_RULE

    if (corparserStruct.CurModule.SupMagicIndexing) {
        //    Go through exported predicate list, and create interface rules.
        for (corExportInfo *export_info = corparserStruct.CurModule.ExportedPreds;
            export_info != NULL; export_info = export_info->next) 
        {
             add_interface_rules(magic_rules, export_info->pred_name,
                export_info->adornment);
        }
    }
}

// *************************************************************************
//    Construct the supplementary magic rules for adrule

void do_supp_magic_on_rule(corClause *adrule, struct corcollection *magic_rules)
{
    int    i,j, num_base_lits;
    int tmp_j;
    int inc;
    corClause *magicrule;
    corLiteral *pred;
    corLiteral *newhead;
    corLiteral *prev_sup;

    rule_num++;
    newhead = build_pred(adrule, adrule->head,0);

    for (j=0;;) {
        if (j < adrule->num_preds && adrule->preds[j]->base_pred)
            j++;
        else
            break;
    }
    num_base_lits = j;
    if (num_base_lits < adrule->num_preds && dont_create_sup0 &&
        !corparserStruct.CurModule.SupMagicIndexing &&
        !corparserStruct.CurModule.UseReturnUnify)
    {
        //    Avoid generating sup_j_0 predicate - in most cases this is an
        //    unnecessary overhead, but is necessary when using supplementary
        //    indexing or return unify.
        prev_sup = build_magic_pred(adrule, adrule->head, 0, 0);
        num_base_lits = 0; 
    }
    else {
        //    Create sup_i_0 predicate.
        magicrule = corcopy_rule(adrule);
        if (!corparserStruct.CurModule.SupMagicIndexing ||
            num_base_lits >= adrule->num_preds)
        {
            coralloc_preds(magicrule, num_base_lits+1);
        }
        else {
            coralloc_preds(magicrule, num_base_lits+2);
        }

        magicrule->preds[0] = build_magic_pred(adrule, adrule->head, 0, 0);
        for (tmp_j=0;tmp_j < num_base_lits; tmp_j++) {
              magicrule->preds[tmp_j+1] = adrule->preds[tmp_j];
        }

        if (num_base_lits >= adrule->num_preds) {
            pred = newhead;
            magicrule->head_deletes_count = adrule->head_deletes_count;
            magicrule->head_deletes = adrule->head_deletes;
        }
        else {
            //    Keep only those vars that either appear in the head or later
            //    in the original rule AND appear in the body of the
            //    supplementary rule that we are going to generate.
            corVarLink *supvars = corbuild_var_list(magicrule->preds[0], 
            corvars_between(adrule, 0, num_base_lits-1, NULL, NULL));
              corVarLink *latervars = corbuild_var_list(newhead,
                vars_from(adrule,num_base_lits));
            supvars = corintersect_var_lists(supvars, latervars);
              pred = build_supmagic(adrule, rule_num, 0, supvars, num_base_lits);
              if (corparserStruct.CurModule.SupMagicIndexing) {
                magicrule->preds[tmp_j+1] = build_goalid(adrule, tmp_j, 1);
            }
        }

        //    WARNING - unnecessary vars should be projected out above
        //    Perhaps all vars are necessary in this case though.
        magicrule->head = prev_sup = pred;
        coradd_rule(magicrule, magic_rules);
        pred->rule = adrule;
    }
  
    for (i=num_base_lits; i < adrule->num_preds;) {
        //    Generate a magic rule for each derived right-hand side predicate
        //    that has an incoming arc in the sip.

        if ( 0 && ! adrule->preds[i]->base_pred && adrule->sips->sip_list ) {
            //    Don't use this for now
            //    WARNING: Eventually just sort the predicates in the sip order,
            //    and then do the sup magic is below.
            if (0) {
            fprintf(stderr,"Error: supmagic cannot deal with sips yet\n");
            struct corsip *s;
            for (s=adrule->sips->sip_list; s; s=s->next)
            if (i == s->head-1) {
                magicrule = corcopy_rule(adrule);
                // rule_num++;
                pred = build_magic_pred(adrule, adrule->preds[i],
                 /* use_share_opts= */ 0, /* argnum = */ i+1);
                magicrule->head = pred;
                if (corparserStruct.CurModule.SupMagicIndexing)
                    create_supmagic_index(magicrule->head);
                  coralloc_preds(magicrule, 1);
                pred->rule = adrule;
                pred = build_supmagic(adrule, rule_num, i+1,
                      corbuild_var_list(prev_sup, NULL),
                      /* derived lit pos.??*/     i);
                magicrule->preds[0] = pred;
            
                magicrule->type = COR_MAGICRULE;
                magicrule->source_rule = adrule;
                magicrule->source_pred = adrule->preds[i];
                coradd_rule(magicrule, magic_rules);
            
                break;    /* only one magic rule per predicate */
            }
          }
        }
    
        if ( !(adrule->preds[i]->base_pred))  {
            magicrule = corcopy_rule(adrule);
            magicrule->head = build_magic_pred(adrule, adrule->preds[i],0,i+1);
            if (corparserStruct.CurModule.SupMagicIndexing)
               create_supmagic_index(magicrule->head);

            if (corparserStruct.CurModule.SupMagicIndexing && add_fail_literal) {
                //    See top of file for explanation.
                coralloc_preds(magicrule,2);
                magicrule->preds[0] = make_fail_literal();
                magicrule->preds[1] = prev_sup;
            }
            else {
                if (0 && corparserStruct.CurModule.NestedSccEval &&
                    prev_sup->kind == COR_SUPMAGIC &&
                    !prev_sup->predicate->ext_magic) {
                  coralloc_preds(magicrule, prev_sup->supp_rule->num_preds);
                  // Added by Ken: copy the body of the rule of the prev_sup to this rule
                  for (int pos = 0; pos < prev_sup->supp_rule->num_preds; pos++)
                    magicrule->preds[pos] = prev_sup->supp_rule->preds[pos];
                }
                else {
                  coralloc_preds(magicrule, 1);
                  magicrule->preds[0] = prev_sup;
                }

                magicrule->head->rule = adrule;
            }

            magicrule->type = COR_MAGICRULE;
            magicrule->source_rule = adrule;
            magicrule->source_pred = adrule->preds[i];

            if (magicrule->source_pred->negated ||
                corhas_grouping_args(adrule->head->args))
            {
                if (!magicrule->clause_names) {
                    // need to create a new name
                    magicrule->clause_names = new corClauseName();
                    magicrule->clause_names->name = corDontCareSymbol;
                    magicrule->clause_names->next = NULL; // is this needed ???
                }
                magicrule->clause_names->negated = 1;
            }
            coradd_rule(magicrule, magic_rules);
        }

        // Generate supplementary rule/ final rule.
        for (j=0;;) {
            if (i+j+1 < adrule->num_preds && adrule->preds[i+j+1]->base_pred)
                j++;
            else break;
        }
        num_base_lits = j;

        magicrule = corcopy_rule(adrule);
        if (i+num_base_lits+1 < adrule->num_preds) {
            //  Keep only those vars that either appear in the head or later in
            //  the original rule AND appear in the body of the supplementary
            //  rule that we are going to generate.
            corVarLink *supvars = corbuild_var_list(prev_sup,
                corvars_between(adrule, i, i+num_base_lits, NULL,NULL));
            corVarLink *latervars = corbuild_var_list(newhead,
                vars_from(adrule,i+num_base_lits+1));
            supvars = corintersect_var_lists(supvars, latervars);
            pred = build_supmagic(adrule, rule_num, i+1,
                supvars,i+num_base_lits+1);
            //    WARNING:  Some space should be freed here.
        }
        else {
            pred = newhead;
            magicrule->head_deletes_count = adrule->head_deletes_count;
            magicrule->head_deletes = adrule->head_deletes;
        }

        int num_deriv_negated = 0;
        if (corparserStruct.CurModule.UseOrdSearch) {
            if (((adrule->preds[i]->negated) || ((pred == newhead) &&
                (corhas_grouping_args(newhead->args)))) && 
                (!corparserStruct.CurModule.Monotonic) &&
                (!adrule->preds[i]->base_pred))
            {
                num_deriv_negated = 1;
            }
        }

        magicrule->head = pred;
        pred->rule = adrule;

        // added by Ken
        if (0 && corparserStruct.CurModule.NestedSccEval &&
            prev_sup->kind == COR_SUPMAGIC &&
            !prev_sup->predicate->ext_magic)
          inc = prev_sup->supp_rule->num_preds - 1;
        else
          inc = 0;

        if (!corparserStruct.CurModule.SupMagicIndexing ||
            !(i+num_base_lits+1 < adrule->num_preds))
        {
                coralloc_preds(magicrule, inc+2+num_base_lits+num_deriv_negated);
        }
        else {
            coralloc_preds(magicrule, inc+2+num_base_lits+num_deriv_negated+1);
        }
        
        magicrule->preds[0] = prev_sup;

        if (0 && corparserStruct.CurModule.NestedSccEval &&
            prev_sup->kind == COR_SUPMAGIC &&
            !prev_sup->predicate->ext_magic)
          // Added by Ken: copy the body of the rule of the prev_sup to this rule
          for (int pos = 0; pos < prev_sup->supp_rule->num_preds; pos++)
            magicrule->preds[pos] = prev_sup->supp_rule->preds[pos];

        if (num_deriv_negated) {
            magicrule->preds[inc+1]=build_magic_pred(adrule,adrule->preds[i],0,i+1);
            magicrule->preds[inc+1]->kind =  COR_DONE_MAGIC;
        }

        magicrule->preds[inc+num_deriv_negated+1] =
            build_pred(adrule, adrule->preds[i], i+1);

        for (tmp_j=0;tmp_j<num_base_lits; tmp_j++) {
            magicrule->preds[inc+tmp_j+num_deriv_negated+2] =
                adrule->preds[i+tmp_j+1];
        }

        if (corparserStruct.CurModule.SupMagicIndexing &&
            (i+num_base_lits+1 < adrule->num_preds))
        {
            magicrule->preds[inc+tmp_j+num_deriv_negated+2] =
                build_goalid(adrule, i+num_base_lits+1, 1);
        }
        i=i+num_base_lits+1;
        
        // If in nested_scc mode, mark the predicate if it will 
        // actually be used in the body of other rules. Added by Ken
        if (0 && corparserStruct.CurModule.NestedSccEval &&
            i < adrule->num_preds &&
            pred->kind == COR_SUPMAGIC &&
            adrule->preds[i]->predicate->scc->scc_no !=
            adrule->head->predicate->scc->scc_no)
          // the next derived predicate is not in the same Scc as head
          // this sup predicate will be used in the body of other rule
          pred->predicate->ext_magic = 1;
          
        pred->supp_rule = magicrule;
        prev_sup = pred;
        if (1 || !corparserStruct.CurModule.NestedSccEval ||
            pred->predicate->ext_magic == 1 || 
            pred->kind != COR_SUPMAGIC)
          coradd_rule(magicrule, magic_rules);
        //generate_index_annotations(magicrule);
        // print_rule(magicrule,stderr);

        /* WARNING::  Create make_index annotation here!! */
    }
}

// *************************************************************************
/*
    This procedure adds new variable names to rule, and stores the number
    of vars at the time it was called in old_num_var_names.
*/

void make_new_rule_vars(corClause *, int)
{
    fprintf(stderr, "make_new_rule_vars not implemented fully yet\n");
}

// *************************************************************************
/*
    This procedure returns the newvarnum'th new rule variable.  The new
    variables must have been created by make_new_rule_vars.
*/

corVarArg* new_rule_var(corClause *, int)
{
    fprintf(stderr, "new_rule_var not implemented fully yet\n");
    return NULL;
}

// *************************************************************************
corVarArg* make_supindex_var(corClause *adrule, int varnum, char letter)
{
    char n[30];
    int varpos;

    sprintf(n,"_%c%d", letter, varnum);
    Name name = corEnterSymbol(n);
    for (varpos = 0; varpos < adrule->num_var_names; varpos++) {
        if (adrule->var_names[varpos] == name)
            break;
    }
    if (varpos >= adrule->num_var_names) {
        adrule->num_var_names++;
        adrule->var_names = (Name *) realloc(adrule->var_names,  
            adrule->num_var_names * sizeof(Name) );
    }
    adrule->var_names[adrule->num_var_names-1] = name;
    return new corVarArg(adrule->var_names[adrule->num_var_names-1], varpos);
}

// *************************************************************************
Name new_pred_name(Name name) 
{
    if (!corparserStruct.CurModule.SupMagicIndexing)
        return name;

    char n[MAX_NAME_LENGTH];
    sprintf(n,"s_%s",name->string());
    return corEnterSymbol(n);
}

// *************************************************************************
Name new_pred_name(char *name) 
{
    if (!corparserStruct.CurModule.SupMagicIndexing)
        return corEnterSymbol(name);

    char *n = new char [strlen(name)+10];
    sprintf(n,"s_%s",name);
    Name newname = corEnterSymbol(n);
    delete n;
    return newname;
}

// *************************************************************************
Name new_interface_pred_name(char *name) 
{
    if (!corparserStruct.CurModule.SupMagicIndexing)
        return corEnterSymbol(name);

    //    Name is a hack to deal with return unify - name MUST start with the same
    //    prefix as supplementary literals (currently sup_)
    //    sprintf(n,"m_i_s_%s",name);
    char *n = new char [strlen(name)+10];
    sprintf(n,"sup_i_s_%s",name);
    Name newname = corEnterSymbol(n);
    delete n;
    return newname;
}

// *************************************************************************
//    Construct an arg list containing only the bound args from the given pred.

corLiteral* build_magic_pred(corClause *adrule, corLiteral * pred,
    int use_share_opt, int argnum)
{
    int    i;
    corCArg *arg_ptr;
    corLiteral *new_pred;
    int count = 0; // count of bound args in pred
    int new_rule_var_count = 0;

    i = 0;
    FOR_EACH_ARG(a, pred->args) {
        if (pred->adorn.test(i++))
            count++;
    } END_EACH_ARG

    if (corparserStruct.CurModule.SupMagicIndexing) count++;
    new_pred = (corLiteral*) new corLiteral(count);
    // malloc(sizeof(Literal) + count * sizeof(ArgPtr));
    // new_pred->args.set_count(count);
    new_pred->kind = COR_MAGIC;
    arg_ptr = &(new_pred->args[0]);
    if (corparserStruct.CurModule.SupMagicIndexing)
         *arg_ptr++ = make_supindex_var(adrule, argnum);    
    i = 0;
    FOR_EACH_ARG(a, pred->args) {
        if (pred->adorn.test(i++)) {
            if (!use_share_opt || a->kindof() != COR_FUNCTOR)
                *arg_ptr++ = *a;
            else
                *arg_ptr++ = new_rule_var(adrule, new_rule_var_count++);
        }
    } END_EACH_ARG
    new_pred->base_pred = 0;
    new_pred->negated = 0;
    if (corparserStruct.CurModule.SupMagicIndexing)
        new_pred->pred = new_pred_name(pred->pred);
    else
        new_pred->pred = pred->pred;
    new_pred->adorn = pred->adorn;
    return new_pred;
}

// *************************************************************************
corLiteral* build_interface_magic_pred(corClause *adrule, corLiteral * pred,
    int use_share_opt, int argnum)
{
    corLiteral *mlit = build_magic_pred(adrule, pred, use_share_opt, argnum);
    mlit->pred = new_interface_pred_name(pred->pred->string());
    mlit->kind = COR_REGULAR;
    //    So that m_ is not printed before predicate name
    return mlit;
}

// *************************************************************************
/*
    Create a new predicate from pred.  The new predicate will have
    an extra argument if supplementary indexing rewriting is used.
    Create a variable using number ``argnum'' and use it in this 
    extra argument.
*/

static corLiteral* build_pred(corClause *adrule, corLiteral *pred, int argnum)
{
    if (!corparserStruct.CurModule.SupMagicIndexing)
        return pred;

    corCArg *arg_ptr;
    corLiteral *new_pred;
    int count = 0; // count of args in pred

    FOR_EACH_ARG(a, pred->args) {
        count++;
    } END_EACH_ARG

    if (corparserStruct.CurModule.SupMagicIndexing) count++;
    new_pred = (corLiteral*) new corLiteral(count);
    // malloc(sizeof(Literal) + count * sizeof(ArgPtr));
    // new_pred->args.set_count(count);
    new_pred->kind = pred->kind;
    // new_pred->predicate = pred->predicate;
    new_pred->pred = new_pred_name(pred->pred);
    new_pred->adorn = pred->adorn;
    new_pred->base_pred = pred->base_pred;
    new_pred->negated = pred->negated;

    arg_ptr = &(new_pred->args[0]);
    if (corparserStruct.CurModule.SupMagicIndexing)
         *arg_ptr++ = make_supindex_var(adrule, argnum);    
    FOR_EACH_ARG(a, pred->args) {
        *arg_ptr++ = *a;
    } END_EACH_ARG
    return new_pred;
}

// *************************************************************************
/*
    If usemagic is set to 1, it assumes that  adrule->preds[argnum] needs a
    goalid literal made for it.  It adds m_ and s_ prefixes and adds adornment.
    It adds a goal-id arg, and projects out free args.  Typically called with
    literal from original rule.  If usemagic is set to 0, m_ and s_ prefixes
    are not added, it is assumed that  adrule->preds[argnum] is a valid
    prototype, and merely has to have adornment added, and then enclosed
    within a goalid literal.  Typically called with a magic literal that has
    been created.
*/

static corLiteral *build_goalid(corClause *adrule, int argnum, int usemagic)
{
    corCArg *arg_ptr;
    corLiteral *new_pred;
    corFuncArg *new_func;
    int count = 0; // count of args in pred
    static char name[MAX_NAME_LENGTH];

    corLiteral *pred = adrule->preds[argnum];
    Name n;

    if (usemagic) {
        int i=0;
        n = new_pred_name(pred->pred);
        corsprint_magic_name(name, n->string(),&(pred->adorn));
        FOR_EACH_ARG(a, pred->args) {
             if (pred->adorn.test(i++))
             count++;
        } END_EACH_ARG
        count++;    // for goalid variable
    } else {
        FOR_EACH_ARG(a, pred->args) {
            count++;
        } END_EACH_ARG
        // goalid variable is assumed to be present already
        n = pred->pred;
        corsprint_pred_name(name, n->string(),&(pred->adorn));
    }

    new_func = corFuncArg::New(corEnterSymbol(name), count);
    arg_ptr = &(new_func->args[0]);

    if (usemagic) {
        *arg_ptr++ = make_supindex_var(adrule, argnum+1);    
        int i = 0;
        FOR_EACH_ARG(a, pred->args) {
            if (pred->adorn.test(i++)) {
                *arg_ptr++ = *a;
            }
        } END_EACH_ARG
    } else {
        FOR_EACH_ARG(a, pred->args) {
            *arg_ptr++ = *a;
        } END_EACH_ARG
    }

    new_pred = (corLiteral*) new corLiteral(1);
    // malloc(sizeof(Literal) + count * sizeof(ArgPtr));
    // new_pred->args.set_count(count);
    new_pred->args[0] = new_func;
    new_pred->kind = COR_REGULAR;
    new_pred->predicate = corFindPredicate(corparserStruct, corGoalIdOpSymbol,1);
    new_pred->pred = corGoalIdOpSymbol;
    // new_pred->adorn = NULL;
    new_pred->base_pred = 1;
    new_pred->negated = 0;

    return new_pred;
}

// *************************************************************************
int count_bound_functor_args(corLiteral *head) 
{
    int count = 0;
    int i = 0;

    FOR_EACH_ARG(a,head->args) {
        if (head->adorn.test(i++) && a->kindof() == COR_FUNCTOR)
            count++;
    } END_EACH_ARG
    
    return(count);
}

// *************************************************************************
/*
    The following fn is used in the opt_share_functors optimization.
    It creates a literal of the form "Var=term", where the var is 
    specified via newvarnum, and "term" is the argnum'th arg of head.
*/

corLiteral* make_functor_literal(corClause *newrule, corLiteral *head, int argnum,
    int newvarnum)
{
    fprintf(stderr, "make_functor_literal not implemented fully yet\n");

    corLiteral *new_pred; 
    corCArg *arg_ptr;
    new_pred = (corLiteral*)
    malloc(sizeof(corLiteral) + 2 * sizeof(ArgPtr));
    new_pred->kind = COR_REGULAR;
    new_pred->args.set_count(2);
    arg_ptr = &(new_pred->args[0]);

    *arg_ptr++ = new_rule_var(newrule, newvarnum);
    *arg_ptr = head->args[argnum];

    new_pred->base_pred = 1;
    new_pred->negated = 0;
    new_pred->pred = corEquOpSymbol;
    //    new_pred->adorn is not initialized
    return new_pred;
}


// *************************************************************************
/*
    This procedure adds literals of the form var=term, for the
    opt_share_functors optimization
*/

void create_functor_literals(corClause *)
{
    fprintf(stderr, "create_functor_literals not implemented yet\n");
}

// *************************************************************************
/*
    This procedure creates a head literal for the magic rule, and implements
    the opt_share_functors optimization if the global flag opt_share_functors
    is set.
*/

corLiteral* build_head_pred(corClause *newrule, corLiteral *oldhead)
{
    int new_rule_var_count = 0;
    int i=0;
    corLiteral *new_pred; 
    corCArg *arg_ptr;

    if (!opt_share_functors || count_bound_functor_args(oldhead) == 0)
        return(oldhead);

    fprintf(stderr, "build_head_pred not implemented fully yet\n");

    new_pred =
        (corLiteral*)malloc(sizeof(corLiteral) + oldhead->arity() * sizeof(ArgPtr));
    new_pred->kind = COR_REGULAR;
    new_pred->args.set_count(oldhead->arity());
    arg_ptr = &(new_pred->args[0]);

    FOR_EACH_ARG(a, oldhead->args) {
        if ( oldhead->adorn.test(i++) && opt_share_functors &&
            a->kindof() == COR_FUNCTOR)
        {
            *arg_ptr++ = new_rule_var(newrule, new_rule_var_count++);
        }
        else {
            *arg_ptr++ = *a;
        }
    } END_EACH_ARG
    new_pred->base_pred = 0;
    new_pred->negated = 0;
    new_pred->pred = oldhead->pred;
    new_pred->adorn = oldhead->adorn;
    return new_pred;
}

// *************************************************************************
static corVarLink *vars_from(corClause *adrule, int pred)
{
    int i;
    corVarLink *result=NULL;
    for (i=pred; i < adrule->num_preds; i++)
        result = corbuild_var_list(adrule->preds[i], result);
    return result;
}

// *************************************************************************
//    WARNING: The following routine damages the varlists that are passed to it.

static corLiteral* build_supmagic(corClause *adrule, int rule_num, int pos,
    corVarLink *var_list, int next_der_lit)
{
    corLiteral *pred;
    corCArg *arg_ptr;
    register corVarLink *link;
    int var_count = 0;
    for (link = var_list; link != NULL; link = link->next)
        var_count++;
    if (corparserStruct.CurModule.SupMagicIndexing &&
        !(adrule->preds[next_der_lit]->base_pred)) var_count ++;
    static char name[30];
    sprintf(name, "sup_%d_%d", rule_num, pos);
    pred = new corLiteral(corparserStruct, corEnterSymbol(name), var_count);
    // Used to be pred = new Literal(NULL, var_count);
    pred->kind = COR_SUPMAGIC;
    pred->pos = pos;
    arg_ptr = &pred->args[var_count];
    if (corparserStruct.CurModule.SupMagicIndexing &&
        !(adrule->preds[next_der_lit]->base_pred) )
    {
        *--arg_ptr = make_supindex_var(adrule, next_der_lit+1);
        var_count--;
    }
    for (link = var_list; --var_count >= 0; ) {
        ASSERT((link != NULL));
        corVarLink *next = link->next;
        *--arg_ptr = link->arg;
        delete link;
        link = next;
    }
    return pred;
}

// *************************************************************************
void create_supmagic_index(corLiteral *pred)
{
    int arity = pred->args.count();
    char *buf = new char[arity+1];

    buf[0] = 'f';
    for (int i=1; i < arity; i++)
        buf[i] = 'b';

    buf[arity] = '\0';
    
    char *name = new char [strlen(pred->pred->string())+10];
    corsprint_magic_name(name, pred->pred->string(),"");

    coradd_make_index_annotation(corparserStruct.make_index_annotations,
        corEnterSymbol(name), &(pred->adorn), make_arglist1(make_arg(buf)), NULL);
    delete buf;
    delete name;
}

// *************************************************************************
void create_interface_index(corLiteral *pred)
{
    int arity = pred->args.count();

    char *buf = new char[arity+1];
    buf[0] = 'b';
    for (int i=1; i < arity; i++)
        buf[i] = 'f';
    buf[arity] = '\0';
    
    char *name = new char [strlen(pred->pred->string())+10];
    corsprint_pred_name(name, pred->pred->string(),"");

    coradd_make_index_annotation(corparserStruct.make_index_annotations,
        corEnterSymbol(name), &(pred->adorn),make_arglist1(make_arg(buf)), NULL);
    delete buf;
    delete name;
}

// *************************************************************************
corLiteral* make_vararglist_literal(corClause *rule, Name name, int startvarnum,
    int arity)
{
    int i;
    corLiteral *lit = new corLiteral(corparserStruct, name, arity);
    for(i=0; i < arity; i++) {
        lit->args[i] = make_supindex_var(rule, i+startvarnum, 'V');
    }
    return lit;
}

// *************************************************************************
corLiteral* make_vararglist_literal(corClause *rule, char *name,
    int startvarnum, int arity)
{
    return make_vararglist_literal(rule, corEnterSymbol(name), startvarnum, arity);
}

// *************************************************************************
void add_interface_rules(struct corcollection *magic_rules, Name predname,
    corBitVector &adorn)
{
    add_interface_rules(magic_rules, predname->string(), adorn);
}

// *************************************************************************
void add_interface_rules(struct corcollection *magic_rules, char * predname,
    corBitVector &adorn)
{
    corClause *rule, *magicrule;
    Name n;

    int num_bound_args=0;
    for (int i=0; i < adorn.length; i++) {
        if(adorn.test(i)) num_bound_args++;
    }

    //    The rule for initial magic predicate ...
    //    sup_i_s_<pred>(X1,..,Xn) :-
    //        m_<pred>(X1,..,Xn), goal_id(m_<pred>(X1,..,Xn.)).

    rule = corcopy_rule(NULL);
    // Dummy so realloc is not confused
    rule->var_names = (Name *) malloc(sizeof(Name));
    coralloc_preds(rule, 2);

    //    Lit is a temporary literal to work off of.
    n = corEnterSymbol(predname);
    corLiteral *lit = make_vararglist_literal(rule, n, 1, adorn.length);
    lit->adorn = adorn;

    rule->head  = build_interface_magic_pred(rule, lit, 0, 1);
    create_interface_index(rule->head);

    //    Hack to meet the interface requirements of build_goalid.
    rule->preds[0] = lit;
    rule->preds[1] =  build_goalid(rule,0,1);

    //    Warning - preds[0] is assigned twice.  Make sure the order is preserved
    int indexingstat = corparserStruct.CurModule.SupMagicIndexing;

    //    To fool system into creating ordinary magic literal, without
    //    supplementary indexing fields.
    corparserStruct.CurModule.SupMagicIndexing = 0;
    rule->preds[0] = build_magic_pred(rule, lit, 0, 1);
    corparserStruct.CurModule.SupMagicIndexing = indexingstat;

    magicrule = rule;
    coradd_rule(magicrule, magic_rules);

    //    Now the rule to create m_s_..(...) facts from m_i_s_..(...) facts
    //    m_s_<pred>(X1,..,Xn) :- sup_i_s_<pred>(X1,..,Xn)

    rule = corcopy_rule(NULL);
    //    Dummy so realloc is not confused
    rule->var_names = (Name *) malloc(sizeof(Name));  
    coralloc_preds(rule, 1);

    if (add_fail_literal) {
        //    See top of file for explanation
        coralloc_preds(rule,2);
        rule->preds[0] = make_fail_literal();
    }
    else {
        coralloc_preds(rule, 1);
    }
    lit = make_vararglist_literal(rule, n, 1, adorn.length);
    lit->adorn = adorn;

    rule->head = build_magic_pred(rule, lit, 0, 1);

    if (corparserStruct.CurModule.SupMagicIndexing)
        create_supmagic_index(rule->head);

    if (add_fail_literal) {
        rule->preds[1] = build_interface_magic_pred(rule, lit, 0, 1);
    }
    else {
        rule->preds[0] = build_interface_magic_pred(rule, lit, 0, 1);
    }
    coradd_rule(rule, magic_rules);

    //    And now the rule for exported predicate...
    //        p(..) :- sup_i_s_(..), s_p(..).

    rule = corcopy_rule(NULL);
    //    Dummy so realloc is not confused
    rule->var_names = (Name *) malloc(sizeof(Name));  
    coralloc_preds(rule, 2);

    rule->head = make_vararglist_literal(rule, predname, 1, adorn.length);
    rule->head->adorn = adorn;

    rule->preds[0] = build_interface_magic_pred(rule, rule->head, 0, 1);
    //    Replace variables other than goalid variable by _ .
    for (int j=1;i < rule->preds[0]->args.count(); i++) {
        rule->preds[0]->args[j] = &corTheDontCareArg;
    }

    rule->preds[1] = build_pred(rule, rule->head, /* argnum=*/ 1);
    coradd_rule(rule, magic_rules);
}

// *************************************************************************
static corLiteral* make_fail_literal()
{
    corLiteral *pred;

    pred = new corLiteral(corparserStruct, corFailSymbol, 0);
    pred->kind = COR_REGULAR;
    pred->base_pred = 1;
    return pred;
}

// *************************************************************************

