/*****************************************************************************/
/*    Token.C   */
/*****************************************************************************/

#include <string.h>
#include <stdlib.h>

#include "Lexer.h"
#include "Macro.h"

/*****************************************************************************/

Token   *next_tokn = (Token *) NULL;
Token    save_tokn;

extern char *tok_sym[];

/*****************************************************************************/

int get_token(Token *tokn)
{
    int ret;

    if (next_tokn){
        if (tokn){
            ret = tokn->kind = next_tokn->kind;
            if (next_tokn->val.str)
                tokn->val.str = next_tokn->val.str;
        }
        next_tokn = (Token *) NULL;
        return(save_tokn.kind);
    }

    if (tokn)
        ret = tokn->kind = ttlex(tokn);
    else 
        ret = save_tokn.kind = ttlex(&save_tokn);

    return(ret);
}

/*****************************************************************************/

int unget_token(Token *tokn)
{
    if (next_tokn){
        fprintf(stderr,"Can only unget one token.\n");
        return(0);
    }

    if (tokn){
        save_tokn.kind = tokn->kind;
        save_tokn.val.str = tokn->val.str;
    }

    next_tokn = &save_tokn;

    return(1);
}

/*****************************************************************************/

/* Use this when you change your mind and want more info about a token 
   read using get_token() - This ONLY works if get_token() was called
   with no argument.  */

Token *get_last_token()
{
    return(&save_tokn);
}

/*****************************************************************************/
void free_token(Token *tokn)
{
    if (!tokn)
        return;

    if ((tokn->kind == SYMBOL) || (tokn->kind == VAR))
        free(tokn->val.str);

    if (tokn != &save_tokn)
        free(tokn);
}

/*****************************************************************************/
//    Returns 1 for variable, 0 for expr

static int check_var_or_expr(char *buf) 
{
    int i = 0;
    int c; 
    
    if (buf == NULL)
        return 0;
    
    c = buf[0];
    if (('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || (c == '_') ||
        (c == '$'))
    {
        while ((c = buf[++i]) != '\0') {
            if (('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') ||
                ('0' <= c && c <= '9') || c == '_')
            {
            }
            else
                return 0;
        }
        return 1;
    }
    else 
        return 0;
}

/*****************************************************************************/
//    Return 1 for coral variable.  Otherwise return 0.

static int coral_var(char *buf)
{
    int i = 0;
    int c; 
    
    if (buf == NULL)
        return 0;
    
    c = buf[0];
    if (c == '$') {
        while ((c = buf[++i]) != '\0') {
            if (('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') ||
                ('0' <= c && c <= '9') || c == '_')
            {
            }
            else
                return 0;
        }
        return 1;
    }
    else 
        return 0;
}

/*****************************************************************************/
/*
    Returns the next expression (argument) whose end is detected by the
    presence of a comma or closing parenthesis.  It takes into account
    comments, nested brackets etc.
*/

Token *get_arg()
{
    int current;
    Token *t;
    int max = 512;
    char *buf = new char[max];
    int i = 0;
    int num_of_paren = 0;        //    number of (
    int num_of_bracket = 0;      //    number of [
    int num_of_braces = 0;       //    number of curly paren
  
    fprintf(stderr, "In get_arg\n");
 
    t = new Token;
    if ((buf == NULL) || (t == NULL)){
        fprintf(stderr,"new failed in get_arg\n");
        exit(1);
    }

    buf[0] = '\0';

    while (1) {
        current = get_token(t);

        fprintf(stderr,"Token is %d, '%s'\n", current,tok_sym[current]);

        switch (current) {

        case COMMA:
            //    CHECK THIS: DO WE REALLY WANT TO FALL THROUGH?
        case RPAREN:
            if (i == 0)
                return (t);

            if ((num_of_paren <= 0) && (num_of_bracket <= 0) &&
                (num_of_braces <= 0))
            {
                buf [i] = '\0';
  
                int j = check_var_or_expr(buf);              
                if (check_var_or_expr(buf)) {
                    //    It is a variable
                    if (coral_var(buf)){
                        t->kind = VAR;
                        t->val.str = strdup(buf);
                    }
                    else {
                        t->kind = IDENT;
                        t->val.str = strdup(buf);
                    }
                }
                else {
                    t->kind = EXPR;
                    t->val.str = strdup(buf);
                }
                delete [] buf;
                return(t);
            }
            buf[i++] = tok_sym[t->kind][0];
            if (current == RPAREN)
                num_of_paren--;
            break;
        case LPAREN:
            num_of_paren++;
            buf[i++] = tok_sym[t->kind][0];
            break;
        case LBRACE:
            num_of_braces++;
            buf[i++] = tok_sym[t->kind][0];
            break;
        case LBRACKET:
            num_of_bracket++;
            buf[i++] = tok_sym[t->kind][0];
            break;
        case RBRACE:
            num_of_braces--;
            buf[i++] = tok_sym[t->kind][0];
            break;
        case RBRACKET:
            num_of_bracket--;
            buf[i++] = tok_sym[t->kind][0];
            break;
        case DOLLAR:
            buf[i++] = tok_sym[t->kind][0];
            break;
        case SYMBOL: 
            strcpy(&(buf[i]),t->val.str);
            i += strlen(t->val.str);
            break;
        case STRING: 
        case IDENT: 
            strcpy(&(buf[i]),t->val.str);
            i += strlen(t->val.str);
            break;
        default:
            strcpy(&(buf[i]),t->val.str);
            i += strlen(t->val.str);
            break;
        }

        if (i+32 >= max) {
            char *new_buf = new (char [max * 2]);
                
            if (new_buf == NULL){
                fprintf(stderr,"new failed in scan_id_keyword\n");
                exit(1);
            }
            strncpy(new_buf, buf, max);
            delete buf;
            buf = new_buf;
            max *= 2;
        }
    }
}

/*****************************************************************************/
