// *************************************************************************
//    Translate.C
// *************************************************************************

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "Translate.h"
#include "HandleExport.h"
#include "MyString.h"
#include "HashDefines.h"
#include "OutputExport.h"
#include "Macro.h"
#include "Lexer.h"
#include "Token.h"
#include "Main.h"

// *************************************************************************

int nesting_level;

struct vars_list *tuple_var_list;
struct vars_list *relation_var_list;

int coral_var_number;

char *coral_file_name = NULL;
FILE *coral_file;

int coral_file_used = 0;

//    For all the export declarations, only one dummy fn is generated.  For
//    every export decl, the code to enter the relation into the database is
//    entered here.  At the end, this string is dumped into the output file
mystring* export_dummy_fn;

//    Export declaration generates a solver for the function in the export
//    decl.  This generates an extern declaration for the function to
//    resolve any conflicts.
mystring* export_header_decl;

//    Declaration of structures used to to execute each 'blob' of CORAL code.
mystring* exec_array_decl;

//    Initialization code for each execution structure to be used.
mystring* exec_array_initializations;

mystring* exec_array_init_calls;

mystring* coral_file_dump_fn;

char* exec_array_init_fn_name;
char* coral_file_dump_fn_name;
char* dummy_fn_name;        
char* dummy_temp_name;

int first_export_decl = 1;
int first_make_tuple = 1;

/*****************************************************************************/

static kind handle_type_declaration(mystring *export_header_decl);
static int get_optional_pointer();
static char *handle_name_declaration();
static void handle_arglist_declaration(handle_export *export_info,
        mystring *export_header_decl);

/*****************************************************************************/
/*    Symbols for all tokens:   */

        /* These need to be correspond to the values in the enumerated
        type tok_type in Lexer.h */

char *tok_sym[] = {

        /* These should not be printed */
    "_EOF_",
    "Symbol",
    "Integer",
    "Double",
    "Variable",
    "Quoted String",
    "Expression",

    "\\[",
    "\\]",
    "init_coral",
    "_coral_export",

    "Identifier",
    ";",

        /* True Coral Tokens */

    "+=",
    "*=",
    ":=",
    "-=",
    "<-",

    "(",
    ")",
    ",",
    ":",

    "//",
    ".",
    "{",
    "}",

    "==>",
    "sips",
    "[",
    "]",

    "_",
    "|",
    "%",

    "~",
    "?",
    "$",
    "true",
    "sum",

    "del",
    "c_module",
    "end_c_module",

    "module",
    "export",
    "alias",
    "end_module",

    "+",
    "-",
    "*",
    "/",

    "=",
    "<>",
    "<",
    ">",
    "<=",
    ">=",

    "<<",
    ">>",
    "bitxor",
    "bitand",
    "bitnot",
    "bitor",

    "mod",
    "@",
    "!",

    "Op",
    "Relop",

    "EOLN (end-of-line)",

    "char",
    "int",
    "float",
    "double",
    "void",
    "string",

    "arg",
    "class",
    "struct",

    "short",
    "long",
    "signed",
    "unsigned",

        /* Misc */
    "as"

};

/*****************************************************************************/

int get_var(infoStruct &info)
{
    char  buf[5000], num_buf[32];
    char *buf_ptr = buf;
    int tok_type;
    int got_type = 0;
    int consults = 0;
    Token tokn;

#define CUR_TYPE info.type_array[info.no_of_params]
#define CUR_VARNAME info.var_names[info.no_of_params]

    CUR_TYPE = INVALID_T;

    while (tok_type = get_token(&tokn)){

        if (tok_type == my_CORAL_END)
            break;

        switch(tok_type){

        case STRING:
            char *tmp_ptr;
            for (tmp_ptr=tokn.val.str; *tmp_ptr; tmp_ptr++)
                if (*tmp_ptr == '\\'){
                    strcpy(buf_ptr,"\\\\");
                    buf_ptr += 2;
                } else if (*tmp_ptr == '"'){
                    strcpy(buf_ptr,"\\\"");
                    buf_ptr += 2;
                } else {
                   *buf_ptr++ = *tmp_ptr; 
                }
            delete [] tokn.val.str;
            break;

        case CONSULT:
            consults++;
        case VAR:
        case SYMBOL:
            strcpy(buf_ptr, tokn.val.str);
            buf_ptr += strlen(tokn.val.str);
            delete [] tokn.val.str;
            break;

        case INUM:
            sprintf(num_buf,"%d",tokn.val.inum);
            strcpy(buf_ptr,num_buf);
            buf_ptr += strlen(num_buf);
            break;

        case RNUM:
            sprintf(num_buf,"%f",tokn.val.rnum);
            strcpy(buf_ptr,num_buf);
            buf_ptr += strlen(num_buf);
            break;

        case DOLLAR:
            tok_type = get_token(&tokn);
#ifdef    DEBUG
            fprintf(stderr,"Got token: %d '%s'.\n",tok_type,tok_sym[tok_type]);
            if ((tok_type == SYMBOL) || (tok_type == VAR))
                fprintf(stderr,"Value of: '%s'.\n",tokn.val.str);
#endif    /* DEBUG */

            got_type = 1;
            switch(tok_type){
            case CHAR:
                CUR_TYPE = CHAR_T;
                break;
            case INT:
                CUR_TYPE = INT_T;
                break;
            case FLOAT:
                CUR_TYPE = FLOAT_T;
                break;
            case DOUBLE:
                CUR_TYPE = DOUBLE_T;
                break;
            case SHORT:
                CUR_TYPE = SHORT_INT;
                break;
            case LONG:
                CUR_TYPE = LONG_INT;
                break;
            case UNSIGNED:
                CUR_TYPE = UNSIGNED_INT;
                break;
            case CHAR_PTR:
                CUR_TYPE = STRING_T;
                break;

            default:  /* A variable */
                got_type = 0;
                *buf_ptr++ = '$';
                if (CUR_TYPE == INVALID_T)
                    CUR_TYPE = INT_T;        /* Type defaults to 'int' */
                CUR_VARNAME = tokn.val.str;
                (info.no_of_params)++;
                CUR_TYPE = INVALID_T;
                break;
            }

            break;

        case EOLN:
            if (buf_ptr != buf){
                coral_file_dump_fn->append("  fprintf(fp, \"%s\\n\", \"");
                coral_file_dump_fn->append(buf);
                coral_file_dump_fn->append("\\n\");\n");
                buf_ptr = buf;
            }
            break;

        case LPAREN:
            tok_type = get_token(&tokn);
            if (tok_type != DOLLAR)
                *buf_ptr++ = '(';
            unget_token(&tokn);
            break;

        case RPAREN:
            if (!got_type)
                *buf_ptr++ = ')';
            got_type = 0;
            break;

        default:
            got_type = 0;
            strcpy(buf_ptr,tok_sym[tok_type]);
            buf_ptr += strlen(tok_sym[tok_type]);
            break;
        }
    }

    if (tok_type != my_CORAL_END)
        fprintf(stderr,"EOF reached in middle of embedded coral section.\n");

    return(consults);

#undef CUR_TYPE
#undef CUR_VARNAME

}

/*****************************************************************************/

/*****************************************************************************/
void init(char *error_msgs_file, char *coral_f) 
{
    char buf[256];
    nesting_level = 0;
    tuple_var_list = NULL;
    relation_var_list = NULL;
    coral_var_number = 0;
    
    //    Setting the error messages file.
    if (error_msgs_file && ((error_file = fopen(error_msgs_file,"w")) != NULL)){

    } else {
        if (error_msgs_file)
            fprintf(stderr, "Cannot open file: %s\n", error_msgs_file);
        error_file = stderr;
        error_msgs_file = strdup("stderr");
    }
    
    //    Setting the file where the coral code will be dumped.
    coral_file_name = STRDUP(coral_f);
    
    exec_array_decl = new mystring;
    exec_array_initializations = new mystring;

    coral_file_dump_fn = new mystring;
    coral_file_dump_fn_name = new_name();
    sprintf(buf, "static void %s()\n{\n", coral_file_dump_fn_name);
    coral_file_dump_fn->append(buf);
    sprintf(buf, "  FILE *fp = fopen(\"%s\", \"w\");\n", coral_file_name);
    coral_file_dump_fn->append(buf);

    exec_array_init_calls = new mystring;
    exec_array_init_fn_name = new_name();
    sprintf(buf, "static void %s()\n{\n", exec_array_init_fn_name);
    exec_array_init_calls->append(buf);
    sprintf(buf, "  %s();\n", coral_file_dump_fn_name);
    exec_array_init_calls->append(buf);    
    sprintf(buf, "  corconsultFile(\"%s\", \"%s\");\n", coral_file_name,
        coral_file_name);
    exec_array_init_calls->append(buf);
    sprintf(buf, "  unlink(\"%s\");\n", coral_file_name);
    exec_array_init_calls->append(buf);

    char *temp_str = new_name();
    dummy_fn_name = new char[strlen(temp_str) + 25];
    strcpy(dummy_fn_name, DUMMY_FN_NAME);
    strcat(dummy_fn_name, temp_str);
    delete temp_str;

    export_dummy_fn = new mystring;
    dummy_temp_name = new_name();
    sprintf(buf, "static void %s()\n{\n   corBuiltinUserTupleRelation *%s;\n",
        dummy_fn_name, dummy_temp_name);
    export_dummy_fn->append(buf);
    
    FILE *temp = output_file;

    //    This statment generates #include <stdio.h> in the module, if it
    //    appears, it is ignored.
    fprintf(temp, "#include <stdio.h>\n");
    fprintf(temp, "#include <string.h>\n");
    fprintf(temp, "#include \"%s\"\n", INCLUDES_FILE);
    fprintf(temp, "\n\n");

    output_init_export(INCLUDES_FILE, dummy_fn_name);
    fprintf(temp, "static void %s();\n\n", exec_array_init_fn_name);
    fprintf(temp,"static %s %s = 0;\n",CORAL_FAIL_VAR_TYPE,
        CORAL_FAIL_VAR);
    line_directive(temp);
}

/*****************************************************************************/
void close(mystring *export_dummy_fn) 
{
    fputs("\n",output_file);
    exec_array_decl->print(output_file);
    delete exec_array_decl;
    exec_array_initializations->print(output_file);
    delete exec_array_initializations;
    exec_array_init_calls->append("}\n");
    coral_file_dump_fn->append("  fclose(fp);\n}\n");
    if (coral_file_used) {
        coral_file_dump_fn->print(output_file);
        exec_array_init_calls->print(output_file);
    }
    delete exec_array_init_calls;
    export_dummy_fn->append("}\n");
    export_dummy_fn->print(output_file);
    delete export_dummy_fn;
}

/*****************************************************************************/
/*
    This function parses an export declaration and stores the necessary
    information in the structure handle_export.  This information is used
    later to output the approprite solver.
*/

void handle_export_declaration(handle_export *export_info,
        mystring *export_header_decl) 
{
    int current;
    Token tokn;

    export_info->set_return_type(handle_type_declaration(export_header_decl));
    
    char *s = handle_name_declaration();
    if (s) {
        export_info->set_name(s);
        export_header_decl->append(s);
    }

    handle_arglist_declaration(export_info, export_header_decl);
    
    current = get_token(&tokn);
    while (1) {
        switch (current) {
        case EOF_TOK:
            fprintf(stderr, "syntax error - unexpected eof\n");
            return ;        
        case SEMICOLON:
            return;
        case AS:
            s = handle_name_declaration(); 
            if (s) export_info->set_exported_name(s);
            current = get_token(&tokn);
            if (current != SEMICOLON) { 
                fprintf(stderr, "; expected \n");
                unget_token(&tokn);
                return;
            }
            return;
        case EOLN:
            break;
        default:
            fprintf(stderr, "; or 'as' expected \n");
            unget_token(&tokn);
            return;
        }
        current = get_token(&tokn);
    }
}

/*****************************************************************************/
static kind handle_type_declaration(mystring *export_header_decl) 
{
    int   current = 0;
    kind  ret_kind;
    Token tokn;
    
    ret_kind.pointer = 0;
    ret_kind.name = NULL;
    ret_kind.type = NULL;
    ret_kind.argname = NULL;
    ret_kind.make = NULL;
    ret_kind._typedef = NULL;
    ret_kind.opt_neg_sign = 0;
    ret_kind.init_pointer = 1;
   
    if ((current = get_token()) == MINUS){
        ret_kind.opt_neg_sign = 1;
        current = get_token();
    }


    ret_kind.value = INVALID_T;
    int done = 0;

    while (!done){
        switch (current){
        
        case EOF_TOK:
            fprintf(stderr, "syntax error in export declaration \n");
            return(ret_kind);
    
        case CHAR:    /* signed/unsigned char not currently allowed */
            ret_kind.value = CHAR_T;
            export_header_decl->append("char ");
            done = 1;
            break;
    
        case FLOAT:
            ret_kind.value = FLOAT_T;
            export_header_decl->append("float ");
            done = 1;
            break;
    
        case DOUBLE:
            ret_kind.value = DOUBLE_T;
            export_header_decl->append("double ");
            done = 1;
            break;
    
        case VOID:
            ret_kind.value = VOID_T;
            export_header_decl->append("void ");
            done = 1;
            break;
    
        case ARG:
            ret_kind.value = ARG_T;
            export_header_decl->append("corArg");
            done = 1;
            break;

        case CHAR_PTR:
            ret_kind.value = STRING_T;
            export_header_decl->append("char *");
            /*  ret_kind.name = "string";  */
            /*  get_token(&tokn);  */
            /*  ret_kind.name = tokn.val.str;  */
            done = 1;
            break;
    
        case CLASS:
            fprintf(stderr, "class declaration in export \
                declaration not handled properly.\n");
            ret_kind.value = CLASS_T;
            //    Get the name
            //    Should put some checks here
            get_token(&tokn);
            ret_kind.name = tokn.val.str;
            done = 1;
            break;
    
        case STRUCT:
            fprintf(error_file, "struct declaration in export \
            declaration not handled properly \n");
            ret_kind.value = STRUCT_T;
            //    Get the name
            get_token(&tokn);
            ret_kind.name = tokn.val.str;
            done = 1;
            break;
    
        case INT:
            if (ret_kind.value == INVALID_T)
                ret_kind.value = INT_T;
            else
                ret_kind.value |= INT_MSK;
            break;
    
        case SHORT:
            if (ret_kind.value == INVALID_T)
                ret_kind.value = SHORT_INT;
            else
                ret_kind.value |= SHORT_MSK;
            fprintf(stdout,"Found short\n");
            break;
    
        case LONG:
            if (ret_kind.value == INVALID_T)
                ret_kind.value = LONG_INT;
            else
                ret_kind.value |= LONG_MSK;
            fprintf(stdout,"Found long\n");
            break;
    
        case SIGNED:    /* 'signed' assumes an integer */
            if (ret_kind.value == INVALID_T)
                ret_kind.value = INT_T;
            else
                ret_kind.value = (ret_kind.value & SIGNED_MSK) | INT_MSK;
            fprintf(stdout,"Found signed\n");
            break;
    
        case UNSIGNED:  /* 'unsigned' assumes an integer */
            if (ret_kind.value == INVALID_T)
                ret_kind.value = UNSIGNED_INT;
            else
                ret_kind.value |= UNSIGNED_MSK;
            fprintf(stdout,"Found unsigned\n");
            break;
    
        default:
            if (ret_kind.value == INVALID_T){
                Token *tmp_tokn = get_last_token();
                done = 1;
                fprintf(error_file,
                   "unimplemented features eg  typedef, macros etc. %d, %s\n",
                    tmp_tokn->kind, tmp_tokn->val.str);
            } else {
                Token *tmp_tokn = get_last_token();
                done = 1;
                unget_token();
            }
            break;
        }

        if (!done)
            current = get_token();
    }


    if (ret_kind.value & INT_T){

        if ((ret_kind.value & SHORT_MSK) && (ret_kind.value & LONG_MSK)){
            fprintf(stderr, "Both long and short specified for type\n");
            return(ret_kind);
        }

        if (ret_kind.value & UNSIGNED_MSK)
            export_header_decl->append("unsigned ");

        if (ret_kind.value & LONG_MSK)
            export_header_decl->append("long ");

        if (ret_kind.value & SHORT_MSK)
            export_header_decl->append("short ");

        export_header_decl->append("int ");
    }

    ret_kind.pointer = get_optional_pointer();
    
    for (int t = 0; t < ret_kind.pointer; t++) {
        export_header_decl->append("*");
    }

    return(ret_kind);
}

/*****************************************************************************/
static int get_optional_pointer() 
{
    int   num = 0;
    Token cur_tokn;

    while (get_token() == STAR)
        num++;

    unget_token();
    return(num);
}

/*****************************************************************************/
static char* handle_name_declaration() 
{
    int   current = 0;
    Token tokn;
   
    current = get_token(&tokn);
 
    switch (current) {
    case EOF_TOK:
        fprintf(stderr, "syntax error in export declaration.\n");
        return(NULL);    

    case IDENT:
        return(tokn.val.str);

    default:
        fprintf(stderr, "syntax error - identifier expected.\n");
        break;
    }

    return(NULL);
}

/*****************************************************************************/
static void handle_arglist_declaration(handle_export *export_info,
        mystring *export_header_decl) 
{
    Token tokn;
    int current;

    if ((current = get_token(&tokn)) != LPAREN) {
        fprintf(stderr, " syntax error, ( expected.\n");
        unget_token(&tokn);
    }
    export_header_decl->append("(");
    
    current = get_token(&tokn);
    
    while(1) {
        switch (current) {
        case EOF_TOK:
            fprintf(stderr, " syntax error- unexpected eof\n");
            return;
        case RPAREN:
            //fprintf(stderr, "done with arglist \n");
            export_header_decl->append(");\n");
            return;
        case SEMICOLON:
            fprintf(stderr, "syntax error - ) expected \n");
            unget_token(&tokn);
            return;
        case COMMA:
            current = get_token(&tokn);
            export_header_decl->append(",");
            break;
        case AS:
            fprintf(stderr, "syntax error - ) expected \n");
            unget_token(&tokn);
            return;
        default:
            unget_token(&tokn);
            export_info->put_next_arg(
                    handle_type_declaration(export_header_decl));
            current = get_token(&tokn);
        }
    }
}

/*****************************************************************************/
//    Routine to search for embedded tokens.

int handle_embedded_token(int current_position, int end_position, char *str)
{
    Token tokn;
    int current;

    current = get_token(&tokn);
    
    fprintf(stderr, "In handle_embedded_token\n");
/* ** I dont understand what this does --- PRAVEEN

    while(((current_token->token_type()) != EOLN) &&
        ((current_token->token_type()) != EOF_TOK))
    {
        switch(current_token->token_type()) {
        case KW_MAKE_TUPLE:
            {
                handle_make_tuple(pos,end_position,str); 
                return 1;
            }
        default:
            pos = input_buffer->get_line_position();
            current_token = input_buffer->get_token();
            break;
        }
    }
*/
    while ((current != EOLN) && (current != EOF_TOK)) {
        current = get_token();
    }

    return(0);
}

/*****************************************************************************/
/*
    For each line between \[ and \], call get_var which generates the
    appropriate code for that line.
*/

int handle_coral_code(infoStruct& info)
{
    int consults = 0;
    coral_file_used = 1;

    coral_file_dump_fn->append("\n  fprintf(fp, \"c_module.\\n\");\n");
    consults = get_var(info);
    coral_file_dump_fn->append("  fprintf(fp, \"end_c_module.\\n\\n\");\n");
    return(consults);
}

/*****************************************************************************/
void generate_initialization_code(char* fn_name, infoStruct& info)
{

    char buf[256];

#    define ADD_CODE exec_array_initializations->append(buf)

    sprintf(buf, "static void %s(corExecutionList& exlist)\n",fn_name);

    ADD_CODE;
    sprintf(buf, "{\n  // initialization code for exlist\n");
    ADD_CODE;
    sprintf(buf, "  exlist.initialize(%d, %d, NULL);\n", EXEC_LIST_SIZE,
        info.no_of_params);
    ADD_CODE;
    
    sprintf(buf, "  corParserStruct *pstruct; int i = 0;\n");
    ADD_CODE;
    sprintf(buf, "  // call parse on %s\n", coral_file_name);
    ADD_CODE;
    sprintf(buf, "  for(;;i++) {\n");
    ADD_CODE;
    sprintf(buf, "    corparserStruct.rule_list = NULL;\n");
    ADD_CODE;
    sprintf(buf, "    corparseStack->parse(stderr,1);\n");
    ADD_CODE;
    sprintf(buf, "    if (corexEnv->c_module_separator) break;\n");
    ADD_CODE;
    sprintf(buf, "    pstruct = new corParserStruct;\n");
    ADD_CODE;

    sprintf(buf, "    corcopy_structs(pstruct, corparserStruct);\n");
    ADD_CODE;

    sprintf(buf, "  // save corParserStruct in exlist\n");
    ADD_CODE;
    sprintf(buf, "    exlist.add(pstruct);\n  }\n");
    ADD_CODE;
    sprintf(buf, "  corexEnv->c_module_separator = 0;\n}\n");
    ADD_CODE;
}

/*****************************************************************************/
struct vars_list *lookup_tuple_var(char *s, struct vars_list *tuple_list)
{
    if (s == NULL)
        return NULL;
    
    while (tuple_list) {
        if (strcmp(s, tuple_list->string) == 0)    
            return tuple_list;
        tuple_list = tuple_list->next;
    }
    
    return(NULL);
}

/*****************************************************************************/
struct vars_list *lookup_relation_var(char *s, struct vars_list *relation_list)
{
    if (s == NULL)
        return NULL;
    
    while (relation_list) {
        if (strcmp(s, relation_list->string) == 0)    
            return relation_list;
        relation_list = relation_list->next;
    }
    
    return(NULL);
}

/*****************************************************************************/
char *new_name()
{
    static int i = 11;
    char buf[10];
    
    sprintf(buf, "%s%d", "temp__", i++);
    char *s = new char[strlen(buf) + 1];
    strcpy(s, buf);
    return(s);
}

/*****************************************************************************/

