/*****************************************************************************/
//    Translate.h
/*****************************************************************************/

#ifndef _Translate_H_
#define _Translate_H_

#include "MyString.h"
#include "Lexer.h"
#include "HandleExport.h"

/*****************************************************************************/
//    List of args if available.

struct arg_list { 
    Token *arg_type;
    struct arg_list *next;
};

/*****************************************************************************/
//    List of tuple variables.

struct vars_list {
    char *string;        // name
    int type;            // pointer or not
    int level;           // nesting level of defn
    int arg_count;
    arg_list *args;
    struct vars_list *next;
};

/*****************************************************************************/
/*
    Structure that carries info about the CORAL commands generated
    corresponding to one 'blob' of embedded CORAL code.
*/

struct infoStruct {
    int no_of_params;
    type_of_arg type_array[MAX_CORAL_VARIABLES];
    char *var_names[MAX_CORAL_VARIABLES];

    infoStruct() {no_of_params = 0;}
};

/*****************************************************************************/

extern mystring* export_dummy_fn;
extern mystring* export_header_decl;
extern mystring* exec_array_decl;
extern mystring* exec_array_init_calls;

extern char* exec_array_init_fn_name;
extern char* dummy_fn_name;        
extern char* dummy_temp_name;

/*****************************************************************************/

void init(char *error_msgs_file, char *coral_f);
void close(mystring *export_dummy_fn);

void handle_export_declaration(handle_export *export_info,
        mystring *export_header_decl);

int handle_embedded_token(int current_position, int end_position, char *str);

int  handle_coral_code(infoStruct& info);

void generate_initialization_code(char *fn_name, infoStruct& info);

struct vars_list *lookup_tuple_var(char *s, struct vars_list *tuple_list);
struct vars_list *lookup_relation_var(char *s, struct vars_list *relation_list);

char *new_name();

/*****************************************************************************/

#endif _Translate_H_

